use strict;
use warnings;

package VProtect;
use Carp qw/confess/;
use Digest::MD5 qw/md5_base64/;
use POSIX qw/strftime/;

my $debug_messages = 1;   # Set to 1 to show timing information for diagnostic purposes.


sub log_msg {
	my $msg = shift;
	print STDERR $msg . "\n";
}

sub debug_message {
	if ($debug_messages == 0) {
		return;
	}
	my $msg = shift;
	my $time = strftime "%b %d %H:%M:%S", localtime;
	print STDERR $time . ' ' . $msg . "\n";
}

sub vim_get_vm_by_uuid {
	# returns moref for vm located by uuid
	my $uuid = shift;

	my $service_content = Vim::get_service_content();

	my $search_index = Vim::get_view(mo_ref => $service_content->searchIndex);

	# First find the VM by it's "instanceUuid" first
	debug_message("Finding VM by instanceUuid");
	my $vm_moref = $search_index->FindByUuid(
		uuid => $uuid,
		vmSearch => 1,
		instanceUuid => 1
	);

	if (defined $vm_moref) {
		debug_message("Returning VM's MoRef that was found by FindByUuid (using instanceUuid)");
		return $vm_moref;
	}

	debug_message("Finding VM by UUID, since VM was not found by instanceUuid");
	my @vm_moref_arr = $search_index->FindAllByUuid(
					uuid => $uuid,
					vmSearch => 1
					);

	# FindAllByUuid API returns _all_ VMs with the $uuid
	# However, if there is one and only one VM with $uuid, only then return it's MoRef
	# Else, error out
	if (defined $vm_moref_arr[0]->[0]) {
		# There is at least one VM with $uuid
		if (defined $vm_moref_arr[0]->[1]) {
			# There are more than one VMs with the same $uuid as their "config.uuid"
			debug_message("More than 1 VMs with the same UUID found");
			return $vm_moref;	# Return undefined, since more than 1 VMs with the
						# same $uuid as their "config.uuid"
		} else {
			# Exactly one VM with $uuid as it's "config.uuid"
			# Good. Return it's MoRef
			debug_message("Returning VM's MoRef that was found by FindAllByUuid (using uuid)");
			return $vm_moref_arr[0]->[0];
		}
	} else {
		debug_message("No VM with found using FindByUuid as well as FindAllByUuid (failure case)");
		return $vm_moref;	# Return undefined, since not a single VM with $uuid as
					# either it's "config.uuid" or it's "config.instanceUuid"
	}

}

sub vim_get_template_by_uuid {
	# returns moref for a Template located by uuid
	my $template_uuid = shift;

	my $template_moref_object;
	my $template_moref;

	my $begin;
	@$begin = Vim::get_service_content()->rootFolder;

	print STDERR "Before find_entity_views\n";
	my $vm_views = Vim::find_entity_views(view_type => 'VirtualMachine',
						begin_entity => @$begin,
						properties => [ 'config' ]
						);

	debug_message ('  Finding a Template in the inventory');
	foreach(@$vm_views) {
		if (!defined($_->config)) {
			next;
		}

		if ($_->config->instanceUuid eq $template_uuid) {
			$template_moref_object = $_;
			last;
		}
	}
	debug_message ('  Exiting Template finding loop');

	if (defined($template_moref_object)) {
		if ($template_moref_object->config->template eq 0) {
			# An entity with UUID $template_uuid was found, but it is a 'VM'
			# and during a Backup, we can't proceed with it.
			die "Cannot find 'Template' " .$template_moref_object->config->name.
			" . Try refreshing the list of VMs\n";
		} else {
			$template_moref = $template_moref_object->get_property('mo_ref');
		}
	}

	return $template_moref;
}

sub vim_get_vm_disks {
	# parameter: moref for a VM
	# returns:

	my $vm_moref = shift;

	my $service_content = Vim::get_service_content();
	my $search_index = Vim::get_view(mo_ref => $service_content->searchIndex);
}


sub vim_get_virtual_machines {
	# This function only returns view information for enumerating VMs --
	#  it does not provide a full view used for backup/restore.  For that,
	#  use get_vm_information_by_moref.

	my ($uuid) = @_;

	my $vm_views;

	my %filter;
	my $begin;

	debug_message(' begin vim_get_virtual_machines');

	@$begin = Vim::get_service_content()->rootFolder;

	if (defined $uuid) {
		debug_message('  begin find_entity_views (w/ uuid)');

		debug_message('   begin find_entity_view host');
		@$begin = Vim::find_entity_view(
			# Only look under the host whose UUID was specified.
			view_type => 'HostSystem',
			filter => { 'hardware.systemInfo.uuid' => $uuid }
		);

		debug_message('   end find_entity_view host');
		debug_message('   begin find_entity_views vms');

		$vm_views = Vim::find_entity_views(
			view_type => 'VirtualMachine',
			begin_entity => @$begin,
			properties => [
				'capability',   # changeTrackingEnabled
				'config',
				'datastore',
				'resourcePool',
				'summary'
			]
		);
		debug_message('  end find_entity_views');
	
		debug_message(' end vim_get_virtual_machines (w/ uuid)');

		return $vm_views;
	} else {
		$vm_views = Vim::find_entity_views(view_type => 'VirtualMachine',
			begin_entity => @$begin,
			properties => [
				'capability',
				'config',
				'datastore',
				'resourcePool',
				'summary'
			]
		);
	}

	debug_message(' end vim_get_virtual_machines');

	if ($vm_views) { return \@$vm_views; }
	else {
		print STDERR "returning no VMs.\n";
		return 0;
	}
}

sub vim_get_hosts {
	my ($entity, $hostuuid) = @_;

	my $entityViews;
	my $begin;

	@$begin = Vim::get_service_content()->rootFolder;

	if (defined $hostuuid) {
		$entityViews = Vim::find_entity_views(
			view_type => $entity,
			begin_entity => @$begin,
			filter => { "summary.hardware.uuid" => $hostuuid }
		);
	} else {
		$entityViews = Vim::find_entity_views(
			view_type => $entity,
			begin_entity => @$begin
		);
	}

	if ($entityViews) { return \@$entityViews; }
	else {
		print STDERR "returning no hosts.\n";
		return 0;
	}
}

sub vim_get_service_content () {
	my $sc = Vim::get_service_content();
	
	return $sc;
}

# Helper sub-routine to return a view for an object with a particular MoRef
#   - The first argument has to always to be a MoRef
#   - The remaining arguments are the properties of the object that need to be retrieved
#	:NOTE:  If only the first argument is passed, then, the view to the
#		object with ALL the properties will be returned
#
#   - Returns a view to the object. If something is not valid, then, returns an 'undef'
# 	(Caller should handle the 'undef' condition)
sub vim_get_view {

	# By default, all arguments are stored in the @_ variable

	my $get_view_to_object;	# Returning a view to "$_[0]"

	if ( ref($_[0]) ne "ManagedObjectReference") {
		# Caller should handle the 'undef' condititon
		return $get_view_to_object;
	}

	if (scalar(@_) == 1) {
		# No arguments other than the MoRef were passed
		# So, return a view with ALL properties
		$get_view_to_object = Vim::get_view( mo_ref => $_[0] );
	} else {
		my @properties; # Create an array of properties
		my $i = 0;
		foreach (@_) {
			if ($i == 0) {
				# Skip the first element, since it is a MoRef and not a property
				$i = $i + 1;
				next;
			} else {
				push(@properties, $_);
			}
		}

		$get_view_to_object = Vim::get_view( mo_ref => $_[0],
					properties => [ @properties ] );
	}

	return $get_view_to_object;

}

sub append_or_die ($$$) {
	my $xmlelem = shift;
	my $name = shift;
	my $data = shift;

	unless (defined($data)) {
		die "Error: Required field $name was not defined.\n";
	}

	$xmlelem->appendTextChild($name, $data);
}

sub get_nic_backing_type ($$) {
	my $xml_ref = shift;
	my $backing_ref = shift;

	if ($$backing_ref->isa('VirtualEthernetCardNetworkBackingInfo')) {
		append_or_die($$xml_ref, '_backing_type', 'VirtualEthernetCardNetworkBackingInfo');
	} elsif ($$backing_ref->isa('VirtualEthernetCardDistributedVirtualPortBackingInfo')) {
		append_or_die($$xml_ref, '_backing_type', 'VirtualEthernetCardDistributedVirtualPortBackingInfo');
		# these are under backing.port, not backing
		append_or_die($$xml_ref, 'switchUuid', $$backing_ref->{'port'}{'switchUuid'});
	} elsif ($$backing_ref->isa('VirtualSriovEthernetCard')) {
		append_or_die($$xml_ref, '_backing_type', 'VirtualSriovEthernetCard');
	} else {
		die "Could not identify backing type: " . ref($$backing_ref) . "\n" . Dumper($$backing_ref) . "\n";
	}

}

sub gen_hw_hash_string ($) {
	# Serialize a basic hardware description of the VM to a string, then take its md5 checksum.
	# This can be used to detect changes in the VM configuration.

	my $hw_hash_ref = shift;

	my $hw_string = "";

	#print STDERR "Processing hw info...\n";

	# These arrays specify the order of the items in the string.
	# Simple keys are basic VM settings.
	# Complex items contain arrays which many have multiple members, and represent VM hardware.
	#   In the hash string, they are ordered by device key #.
	my @simplekeys = ('version', 'numCPU', 'memoryMB', 'firmware');
	my @complexkeys = ('Cdrom', 'DiskCtrl', 'Floppy', 'NIC', 'Parallel', 'PCI', 'Serial', 'USB', 'USBXHCI', 'VirtualDisk', 'VMCI');

	for my $key (@simplekeys) {
		if (not defined $hw_hash_ref->{$key}) {
			next;
		}

		$hw_string .= $key . ':' . $hw_hash_ref->{$key} . ';';
	}

	for my $key (@complexkeys) {
		if (not defined $hw_hash_ref->{$key}) {
			next;
		}

		my @sorted_list = sort @{$hw_hash_ref->{$key}};
		
		foreach my $item (@sorted_list) {
			$hw_string .= $item . ';';
		}
	}

	print STDERR "Hardware string: " . $hw_string . "\n";
	my $md5hash = md5_base64($hw_string);
	print STDERR "Hash string hash: " . $md5hash . "\n";

	return $md5hash;
}

sub get_vm_information_by_moref ($$$$) {
	# This is similar to get_vm_information, but includes more information, and only
	# operates on one VM.  It is oriented toward supporting backup and restore rather than
	# management operations.

	my $vm_moref = shift;
	my $keys = shift;
	my $xml_doc = shift;
	my $xml_node = shift;

	#my @vm_hw_hash_array = ();  # Information used to detect VM configuration changes
	my $vm_hw_hash_ref = {};

	sub append_if_def {
		my $xmlelem = shift;
		my $name = shift;
		my $data = shift;

		if (defined($data)) {
			$xmlelem->appendTextChild($name, $data);
		}
	}

	debug_message('begin get_view VM');
	my $vm = Vim::get_view(mo_ref => $vm_moref);
	debug_message('end get_view VM');

	my @keys_array_ref = @$keys;

	my $vm_x = $xml_doc->createElement('virtual_machine');

	if (not defined ($vm->config)) {
		die "Unable to retrieve VM configuration. (Config not defined.)\n";
	}

	if (defined($vm->config)) {   # TODO
		# Translate %5c to \ and %2f to /
		my $name = $vm->config->name;
		$name =~ s/%5c/\\/g;
		$name =~ s/%2f/\//g;

		append_if_def($vm_x, 'name', $name);

		append_if_def($vm_x, 'uuid', $vm->config->uuid);

		append_if_def($vm_x, 'version', $vm->config->version);
		$vm_hw_hash_ref->{'version'} = $vm->config->version;

		append_if_def($vm_x, 'changeTrackingEnabled', $vm->config->changeTrackingEnabled);

		append_if_def($vm_x, 'guestId', $vm->config->guestId);
		append_if_def($vm_x, 'guestFullName', $vm->config->guestFullName);
		append_if_def($vm_x, 'alternateGuestName', $vm->config->alternateGuestName);

		append_if_def($vm_x, 'annotation', $vm->config->annotation);

		append_if_def($vm_x, 'cpuHotAddEnabled', $vm->config->cpuHotAddEnabled);
		append_if_def($vm_x, 'cpuHotRemoveEnabled', $vm->config->cpuHotRemoveEnabled);

		append_if_def($vm_x, 'firmware', $vm->config->firmware);
		$vm_hw_hash_ref->{'firmware'} = $vm->config->firmware;

		append_if_def($vm_x, 'maxMksConnections', $vm->config->maxMksConnections);
		append_if_def($vm_x, 'memoryHotAddEnabled', $vm->config->memoryHotAddEnabled);
		append_if_def($vm_x, 'memoryReservationLockedToMax', $vm->config->memoryReservationLockedToMax);

		my $config_x = $xml_doc->createElement('config');
		{
			# configuration information
			# bootOptions
			my $bootoptions_x = $xml_doc->createElement('bootOptions');

			append_if_def($bootoptions_x, 'bootDelay', $vm->config->bootOptions->bootDelay);
			append_if_def($bootoptions_x, 'enterBIOSSetup', $vm->config->bootOptions->enterBIOSSetup);

			# extraConfig
			if (defined $vm->config->extraConfig) {
				my $extraconfig_x = $xml_doc->createElement('extraConfig');

				my $extraConfig = $vm->config->extraConfig;
				foreach my $entry (@$extraConfig) {
					if ($entry->key eq 'cpuid.coresPerSocket') {
						append_if_def($extraconfig_x, 'cpuid.coresPerSocket', $entry->value);
					}
				}

				$config_x->appendChild($extraconfig_x);
			}

			$config_x->appendChild($bootoptions_x);
			$vm_x->appendChild($config_x);
		}


		my $vm_hw = $xml_doc->createElement('hardware');

		append_or_die($vm_hw, 'numCPU', $vm->config->hardware->numCPU);
		$vm_hw_hash_ref->{'numCPU'} = $vm->config->hardware->numCPU;

		append_if_def($vm_hw, 'memoryMB', $vm->config->hardware->memoryMB);
		$vm_hw_hash_ref->{'memoryMB'} = $vm->config->hardware->memoryMB;

		my $device_hw = $xml_doc->createElement('devices');

		my $devices = $vm->config->hardware->device;
		foreach (@$devices) {

			print STDERR "Current Virtual Device is - " . ref($_) . "\n";

			if (ref($_) eq 'ParaVirtualSCSIController') {
				my $pvscsi_x = $xml_doc->createElement('ParaVirtualSCSIController');

				append_if_def($pvscsi_x, 'busNumber',          $_->{'busNumber'});
				append_if_def($pvscsi_x, 'controllerKey',      $_->{'controllerKey'});
				append_if_def($pvscsi_x, 'hotAddRemove',       $_->{'hotAddRemove'});
				append_if_def($pvscsi_x, 'key',                $_->{'key'});
				append_if_def($pvscsi_x, 'scsiCtlrUnitNumber', $_->{'scsiCtlrUnitNumber'});
				append_if_def($pvscsi_x, 'sharedBus',          $_->{'sharedBus'});
				append_if_def($pvscsi_x, 'unitNumber',         $_->{'unitNumber'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');

				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->{'label'});
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->{'summary'});

				$pvscsi_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($pvscsi_x);

				push @{$vm_hw_hash_ref->{'DiskCtrl'}}, $_->{'key'} . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualBusLogicController') {
				my $buslog_x = $xml_doc->createElement('VirtualBusLogicController');

				append_if_def($buslog_x, 'busNumber',          $_->{'busNumber'});
				append_if_def($buslog_x, 'controllerKey',      $_->{'controllerKey'});
				append_if_def($buslog_x, 'hotAddRemove',       $_->{'hotAddRemove'});
				append_if_def($buslog_x, 'key',                $_->{'key'});
				append_if_def($buslog_x, 'scsiCtlrUnitNumber', $_->{'scsiCtlrUnitNumber'});
				append_if_def($buslog_x, 'sharedBus',          $_->{'sharedBus'});
				append_if_def($buslog_x, 'unitNumber',         $_->{'unitNumber'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');

				append_if_def($deviceInfo_x, 'label',   $_->{'label'});
				append_if_def($deviceInfo_x, 'summary', $_->{'summary'});

				$buslog_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($buslog_x);

				push @{$vm_hw_hash_ref->{'DiskCtrl'}}, $_->{'key'} . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualCdrom') {
				my $cdrom_x = $xml_doc->createElement('VirtualCdrom');

				append_if_def($cdrom_x, 'controllerKey', $_->{'controllerKey'});
				append_if_def($cdrom_x, 'key',           $_->{'key'});
				append_if_def($cdrom_x, 'unitNumber',    $_->{'unitNumber'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->{'label'});
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->{'summary'});

				my $connectable_x = $xml_doc->createElement('connectable');

				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->{'allowGuestControl'});
				append_if_def($connectable_x, 'connected',         $_->connectable->{'connected'});
				append_if_def($connectable_x, 'startConnected',    $_->connectable->{'startConnected'});
				append_if_def($connectable_x, 'status',            $_->connectable->{'status'});

				my $backing_x = $xml_doc->createElement('backing');
				if ($_->backing->isa('VirtualCdromIsoBackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualCdromIsoBackingInfo');
				} elsif ($_->backing->isa('VirtualCdromAtapiBackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualCdromAtapiBackingInfo');
				} elsif ($_->backing->isa('VirtualCdromRemotePassthroughBackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualCdromRemotePassthroughBackingInfo');
				} elsif ($_->backing->isa('VirtualCdromRemoteAtapiBackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualCdromRemoteAtapiBackingInfo');
				} else {
					die "Could not identify Cdrom backing type:" . Dumper($_->backing) . "\n";
				}

				append_if_def($backing_x, 'deviceName',    $_->backing->{'deviceName'});
				append_if_def($backing_x, 'exclusive',     $_->backing->{'exclusive'});
				append_if_def($backing_x, 'fileName',      $_->backing->{'fileName'});
				append_if_def($backing_x, 'useAutoDetect', $_->backing->{'useAutoDetect'});

				$cdrom_x->appendChild($backing_x);
				$cdrom_x->appendChild($connectable_x);
				$cdrom_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($cdrom_x);

				push @{$vm_hw_hash_ref->{'Cdrom'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualDisk') {
				if ($_->{'backing'}->isa('VirtualDiskRawDiskMappingVer1BackingInfo')) {
					if ($_->{'backing'}->compatibilityMode eq 'physicalMode') {
						print STDERR "Physical RDM found. Skipping it.\n";
						next;
					}
				}

				my $disk_x = $xml_doc->createElement('VirtualDisk');
				append_if_def($disk_x, 'capacityInKB',  $_->{'capacityInKB'});
				append_if_def($disk_x, 'controllerKey', $_->{'controllerKey'});
				append_or_die($disk_x, 'key',           $_->{'key'});

				# Bug 16541 - Add exclude Information to metadata.xml file
				foreach my $iterator (@keys_array_ref) {
					if (($_->{'key'} eq $iterator) || ($_->{'backing'}->diskMode ne 'persistent')) {
						$disk_x->appendTextChild('is_excluded','1');
					}
				}

				append_if_def($disk_x, 'unitNumber',    $_->{'unitNumber'});

				my $backing_x = $xml_doc->createElement('backing');

				if ($_->{'backing'}->isa('VirtualDiskFlatVer2BackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualDiskFlatVer2BackingInfo');
				} elsif ($_->{'backing'}->isa('VirtualDiskRawDiskMappingVer1BackingInfo')) {
					if ($_->{'backing'}->compatibilityMode eq 'virtualMode') {
						append_if_def($backing_x, '_backing_type', 'VirtualDiskRawDiskMappingVer1BackingInfo');
						append_if_def($backing_x, 'contentId', $_->{'backing'}->contentId);
						append_if_def($backing_x, 'deviceName', $_->{'backing'}->deviceName);
						append_if_def($backing_x, 'lunUuid', $_->{'backing'}->lunUuid);
					} elsif ($_->{'backing'}->compatibilityMode eq 'physicalMode') {
						die "Virtual disks using physical mode raw device mapping (RDM) are not supported.\n";
					} else {
						die "Could not identify compatibility mode for RDM disk\n" . Dumper($_) . "\n";
					}
				} else {
					die "Could not identify backing for disk: " . ref $_->{'backing'} . "\n";
				}

				if (not defined($_->backing->{'uuid'})) {
					my $msg = "The virtual disk with filename" . $_->backing->{'filename'} .
					" does not have a UUID assigned to it.\n" .
					"Please assign a UUID using the command " .
					"\"vmkfstools -J setuuid <path_to_vm>/vm_name.vmdk\" on your ESX host.  (See VMware KB 1033370 for more info.)\n" .
					"Restart the management agents for this change to take effect. (http://kb.vmware.com/kb/1003490)\n" .
					"key " . $_->{'key'} . "\n";
					die $msg;
				}

				append_or_die($backing_x, 'uuid',            $_->backing->{'uuid'});
				append_if_def($backing_x, 'diskMode',        $_->backing->{'diskMode'});
				append_if_def($backing_x, 'fileName',        $_->backing->{'fileName'});
				append_if_def($backing_x, 'thinProvisioned', $_->backing->{'thinProvisioned'});
				append_if_def($backing_x, 'writeThrough',    $_->backing->{'writeThrough'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label', $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$disk_x->appendChild($backing_x);
				$disk_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($disk_x);

				push @{$vm_hw_hash_ref->{'VirtualDisk'}}, $_->key . ':' . ref($_) . ':' . $_->capacityInKB . "KB";
			}
			elsif (ref($_) eq 'VirtualE1000') {
				my $e1000_x = $xml_doc->createElement('VirtualE1000');

				append_if_def($e1000_x, 'addressType',      $_->{'addressType'});
				append_if_def($e1000_x, 'controllerKey',    $_->{'controllerKey'});
				append_if_def($e1000_x, 'key',              $_->{'key'});
				append_if_def($e1000_x, 'macAddress',       $_->{'macAddress'});
				append_if_def($e1000_x, 'unitNumber',       $_->{'unitNumber'});
				append_if_def($e1000_x, 'wakeOnLanEnabled', $_->{'wakeOnLanEnabled'});

				my $backing_x = $xml_doc->createElement('backing');
				get_nic_backing_type(\$backing_x, \$_->backing);
				append_if_def($backing_x, 'deviceName',        $_->backing->{'deviceName'});
				append_if_def($backing_x, 'inPassthroughMode', $_->backing->{'inPassthroughMode'});
				append_if_def($backing_x, 'useAutoDetect',     $_->backing->{'useAutoDetect'});

				my $connectable_x = $xml_doc->createElement('connectable');
				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->{'allowGuestControl'});
				append_if_def($connectable_x, 'connected',         $_->connectable->{'connected'});
				append_if_def($connectable_x, 'startConnected',    $_->connectable->{'startConnected'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$e1000_x->appendChild($backing_x);
				$e1000_x->appendChild($connectable_x);
				$e1000_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($e1000_x);

				push @{$vm_hw_hash_ref->{'NIC'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualE1000e') {
				my $e1000e_x = $xml_doc->createElement('VirtualE1000e');
				
				append_if_def($e1000e_x, 'addressType',      $_->{'addressType'});
				append_if_def($e1000e_x, 'controllerKey',    $_->{'controllerKey'});
				append_if_def($e1000e_x, 'key',              $_->{'key'});
				append_if_def($e1000e_x, 'macAddress',       $_->{'macAddress'});
				append_if_def($e1000e_x, 'unitNumber',       $_->{'unitNumber'});
				append_if_def($e1000e_x, 'wakeOnLanEnabled', $_->{'wakeOnLanEnabled'});

				my $backing_x = $xml_doc->createElement('backing');
				get_nic_backing_type(\$backing_x, \$_->backing);
				append_if_def($backing_x, 'deviceName', $_->backing->{'deviceName'});
				append_if_def($backing_x, 'inPassthroughMode', $_->backing->{'inPassthroughMode'});
				append_if_def($backing_x, 'useAutoDetect',     $_->backing->{'useAutoDetect'});

				my $connectable_x = $xml_doc->createElement('connectable');
				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->{'allowGuestControl'});
				append_if_def($connectable_x, 'connected',         $_->connectable->{'connected'});
				append_if_def($connectable_x, 'startConnected',    $_->connectable->{'startConnected'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$e1000e_x->appendChild($backing_x);
				$e1000e_x->appendChild($connectable_x);
				$e1000e_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($e1000e_x);

				push @{$vm_hw_hash_ref->{'NIC'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualFloppy') {
				my $floppy_x = $xml_doc->createElement('VirtualFloppy');

				append_if_def($floppy_x, 'controllerKey', $_->{'controllerKey'});
				append_if_def($floppy_x, 'key',           $_->{'key'});
				append_if_def($floppy_x, 'unitNumber',    $_->{'unitNumber'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				my $connectable_x = $xml_doc->createElement('connectable');
				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->{'allowGuestControl'});
				append_if_def($connectable_x, 'connected',         $_->connectable->{'connected'});
				append_if_def($connectable_x, 'startConnected',    $_->connectable->{'startConnected'});

				my $backing_x = $xml_doc->createElement('backing');
				append_if_def($backing_x, 'deviceName',    $_->backing->{'deviceName'});
				append_if_def($backing_x, 'useAutoDetect', $_->backing->{'useAutoDetect'});

				if ($_->{'backing'}->isa('VirtualFloppyRemoteDeviceBackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualFloppyRemoteDeviceBackingInfo');
				} elsif ($_->{'backing'}->isa('VirtualFloppyImageBackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualFloppyImageBackingInfo');
					append_if_def($backing_x, 'fileName', $_->backing->{'fileName'});
				} elsif ($_->{'backing'}->isa('VirtualFloppyDeviceBackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualFloppyDeviceBackingInfo');
					append_if_def($backing_x, 'deviceName', $_->backing->{'deviceName'});
				} else {
					confess "Could not identify backing for floppy: " . ref $_->{'backing'};
				}
				
				$floppy_x->appendChild($backing_x);
				$floppy_x->appendChild($deviceInfo_x);
				$floppy_x->appendChild($connectable_x);
				$device_hw->appendChild($floppy_x);

				push @{$vm_hw_hash_ref->{'Floppy'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualHdAudioCard') {
				my $hdaudio_x = $xml_doc->createElement('VirtualHdAudioCard');

				append_if_def($hdaudio_x, 'key', $_->{'key'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				my $connectable_x = $xml_doc->createElement('connectable');
				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->{'allowGuestControl'});
				append_if_def($connectable_x, 'connected', $_->connectable->{'connected'});
				append_if_def($connectable_x, 'startConnected', $_->connectable->{'startConnected'});

				my $backing_x = $xml_doc->createElement('backing');
				if (ref($_->backing) eq 'VirtualSoundCardDeviceBackingInfo') {
					append_if_def($backing_x, '_backing_type', 'VirtualSoundCardDeviceBackingInfo');
				} else {
					die 'Cannot identify backing type for VirtualHdAudioCard' . $_->backing . "\n";
				}
				append_if_def($backing_x, 'deviceName',    $_->backing->{'deviceName'});
				append_if_def($backing_x, 'useAutoDetect', $_->backing->{'useAutoDetect'});

				$hdaudio_x->appendChild($deviceInfo_x);
				$hdaudio_x->appendChild($connectable_x);
				$hdaudio_x->appendChild($backing_x);
				$device_hw->appendChild($hdaudio_x);
			}
			elsif (ref($_) eq 'VirtualEnsoniq1371') {
				my $ensoniq_x = $xml_doc->createElement('VirtualEnsoniq1371');

				append_if_def($ensoniq_x, 'key', $_->{'key'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				my $connectable_x = $xml_doc->createElement('connectable');
				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->{'allowGuestControl'});
				append_if_def($connectable_x, 'connected', $_->connectable->{'connected'});
				append_if_def($connectable_x, 'startConnected', $_->connectable->{'startConnected'});

				my $backing_x = $xml_doc->createElement('backing');
				if (ref($_->backing) eq 'VirtualSoundCardDeviceBackingInfo') {
					append_if_def($backing_x, '_backing_type', 'VirtualSoundCardDeviceBackingInfo');
				} else {
					die 'Cannot identify backing type for VirtualEnsoniq1371' . $_->backing . "\n";
				}
				append_if_def($backing_x, 'deviceName',    $_->backing->{'deviceName'});
				append_if_def($backing_x, 'useAutoDetect', $_->backing->{'useAutoDetect'});

				$ensoniq_x->appendChild($deviceInfo_x);
				$ensoniq_x->appendChild($connectable_x);
				$ensoniq_x->appendChild($backing_x);
				$device_hw->appendChild($ensoniq_x);
			}
			elsif (ref($_) eq 'VirtualSoundBlaster16') {
				my $blaster_x = $xml_doc->createElement('VirtualSoundBlaster16');

				append_if_def($blaster_x, 'key', $_->{'key'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				my $connectable_x = $xml_doc->createElement('connectable');
				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->{'allowGuestControl'});
				append_if_def($connectable_x, 'connected', $_->connectable->{'connected'});
				append_if_def($connectable_x, 'startConnected', $_->connectable->{'startConnected'});

				my $backing_x = $xml_doc->createElement('backing');
				if (ref($_->backing) eq 'VirtualSoundCardDeviceBackingInfo') {
					append_if_def($backing_x, '_backing_type', 'VirtualSoundCardDeviceBackingInfo');
				} else {
					die 'Cannot identify backing type for VirtualSoundBlaster16' . $_->backing . "\n";
				}
				append_if_def($backing_x, 'deviceName',    $_->backing->{'deviceName'});
				append_if_def($backing_x, 'useAutoDetect', $_->backing->{'useAutoDetect'});

				$blaster_x->appendChild($deviceInfo_x);
				$blaster_x->appendChild($connectable_x);
				$blaster_x->appendChild($backing_x);
				$device_hw->appendChild($blaster_x);
			}
			elsif (ref($_) eq 'VirtualKeyboard') {
				# don't back this up.
			}
			elsif (ref($_) eq 'VirtualIDEController') {
				my $ide_x = $xml_doc->createElement('VirtualIDEController');

				append_if_def($ide_x, 'busNumber', $_->{'busNumber'});
				append_if_def($ide_x, 'key',       $_->{'key'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$ide_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($ide_x);

				push @{$vm_hw_hash_ref->{'DiskCtrl'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualLsiLogicController') {
				my $lsi_x = $xml_doc->createElement('VirtualLsiLogicController');

				append_if_def($lsi_x, 'busNumber',          $_->{'busNumber'});
				append_if_def($lsi_x, 'controllerKey',      $_->{'controllerKey'});
				append_if_def($lsi_x, 'hotAddRemove',       $_->{'hotAddRemove'});
				append_if_def($lsi_x, 'key',                $_->{'key'});
				append_if_def($lsi_x, 'scsiCtlrUnitNumber', $_->{'scsiCtlrUnitNumber'});
				#append_if_def($lsi_x, 'sharedBus',          $_->{'sharedBus'});
				append_if_def($lsi_x, 'unitNumber',         $_->{'unitNumber'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				
				append_if_def($deviceInfo_x, 'label',   $_->{'label'});  # TODO: looks wrong, examine in 6.2
				append_if_def($deviceInfo_x, 'summary', $_->{'summary'});

				$lsi_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($lsi_x);

				push @{$vm_hw_hash_ref->{'DiskCtrl'}} => $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualLsiLogicSASController') {
				my $lsi_x = $xml_doc->createElement('VirtualLsiLogicSASController');

				append_if_def($lsi_x, 'busNumber',     $_->{'busNumber'});
				append_if_def($lsi_x, 'controllerKey', $_->{'controllerKey'});
				append_if_def($lsi_x, 'hotAddRemove',  $_->{'hotAddRemove'});
				append_if_def($lsi_x, 'key',           $_->{'key'});
				append_if_def($lsi_x, 'scsiCtlrUnitNumber', $_->{'scsiCtlrUnitNumber'});
				# don't backup "sharedBus" since we don't use it on the restore
				append_if_def($lsi_x, 'unitNumber',    $_->{'unitNumber'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');

				append_if_def($deviceInfo_x, 'label',   $_->{'label'});  # TODO: looks wrong, examine in 6.2
				append_if_def($deviceInfo_x, 'summary', $_->{'summary'});

				$lsi_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($lsi_x);

				push @{$vm_hw_hash_ref->{'DiskCtrl'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualAHCIController') {
				# Supported from vSphere 5.5 and Unitrends Version 7.4
				my $ahci_x = $xml_doc->createElement('VirtualAHCIController');

				append_if_def($ahci_x, 'busNumber',     $_->{'busNumber'});
				append_if_def($ahci_x, 'controllerKey', $_->{'controllerKey'});
				append_if_def($ahci_x, 'key',           $_->{'key'});
				append_if_def($ahci_x, 'unitNumber',    $_->{'unitNumber'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');

				append_if_def($deviceInfo_x, 'label',   $_->{'label'});
				append_if_def($deviceInfo_x, 'summary', $_->{'summary'});

				$ahci_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($ahci_x);

				push @{$vm_hw_hash_ref->{'DiskCtrl'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualMachineVideoCard') {
				# don't back this up
				#my $video_x = $xml_doc->createElement('VirtualMachineVideoCard');

				#append_if_def($video_x, 'controllerKey',    $_->controllerKey);
				#append_if_def($video_x, 'key',              $_->key);
				#append_if_def($video_x, 'unitNumber',       $_->unitNumber);

				#append_if_def($video_x, 'enable3DSupport',  $_->enable3DSupport);
				#append_if_def($video_x, 'numDisplays',      $_->numDisplays);
				#append_if_def($video_x, 'useAutoDetect',    $_->useAutoDetect);
				#append_if_def($video_x, 'videoRamSizeInKB', $_->videoRamSizeInKB);

				#$device_hw->appendChild($video_x);
			}
			elsif (ref($_) eq 'VirtualMachineVMCIDevice') {
				my $vmci_x = $xml_doc->createElement('VirtualMachineVMCIDevice');

				append_if_def($vmci_x, 'allowUnrestrictedCommunication', $_->allowUnrestrictedCommunication);
				append_if_def($vmci_x, 'controllerKey',                  $_->controllerKey);
				append_if_def($vmci_x, 'key',                            $_->key);
				append_if_def($vmci_x, 'unitNumber',                     $_->unitNumber);

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$vmci_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($vmci_x);

				push @{$vm_hw_hash_ref->{'VMCI'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualMachineVMIROM') {
				my $vmi_x = $xml_doc->createElement('VirtualMachineVMIROM');
				$device_hw->appendChild($vmi_x);
			}
			elsif (ref($_) eq 'VirtualParallelPort') {
				my $parport_x = $xml_doc->createElement('VirtualParallelPort');

				append_if_def($parport_x, 'controllerKey',    $_->controllerKey);
				append_if_def($parport_x, 'key',              $_->key);
				append_if_def($parport_x, 'unitNumber',       $_->unitNumber);

				my $backing_x = $xml_doc->createElement('backing');
				if ($_->backing->isa('VirtualParallelPortFileBackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualParallelPortFileBackingInfo');
					append_if_def($backing_x, 'fileName', $_->backing->fileName);
				} elsif ($_->backing->isa('VirtualParallelPortDeviceBackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualParallelPortDeviceBackingInfo');
					append_if_def($backing_x, 'deviceName', $_->backing->deviceName);
					append_if_def($backing_x, 'useAutoDetect', $_->backing->useAutoDetect);
				} else {
					die "Unable to identify parallel port backing type: " . Dumper($_) . "\n";
				}

				my $connectable_x = $xml_doc->createElement('connectable');
				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->allowGuestControl);
				append_if_def($connectable_x, 'connected',         $_->connectable->connected);
				append_if_def($connectable_x, 'startConnected',    $_->connectable->startConnected);
				append_if_def($connectable_x, 'status',            $_->connectable->status);

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$parport_x->appendChild($backing_x);
				$parport_x->appendChild($connectable_x);
				$parport_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($parport_x);

				push @{$vm_hw_hash_ref->{'Parallel'}}, $_->{'key'} . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualPCIController') {
				my $pci_x = $xml_doc->createElement('VirtualPCIController');

				append_if_def($pci_x, 'busNumber', $_->{'busNumber'});
				append_if_def($pci_x, 'key',       $_->{'key'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$pci_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($pci_x);

				push @{$vm_hw_hash_ref->{'PCI'}}, $_->{'key'} . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualPCIPassthrough') {
				# Don't back this up.
			}
			elsif (ref($_) eq 'VirtualPCNet32') {
				my $pcnic_x = $xml_doc->createElement('VirtualPCNet32');
				append_if_def($pcnic_x, 'addressType',      $_->{'addressType'});
				append_if_def($pcnic_x, 'controllerKey',    $_->{'controllerKey'});
				append_if_def($pcnic_x, 'key',              $_->{'key'});
				append_if_def($pcnic_x, 'macAddress',       $_->{'macAddress'});
				append_if_def($pcnic_x, 'unitNumber',       $_->{'unitNumber'});
				append_if_def($pcnic_x, 'wakeOnLanEnabled', $_->{'wakeOnLanEnabled'});

				my $backing_x = $xml_doc->createElement('backing');
				get_nic_backing_type(\$backing_x, \$_->backing);
				append_if_def($backing_x, 'deviceName',        $_->backing->{'deviceName'});
				append_if_def($backing_x, 'inPassthroughMode', $_->backing->{'inPassthroughMode'});
				append_if_def($backing_x, 'useAutoDetect',     $_->backing->{'useAutoDetect'});

				my $connectable_x = $xml_doc->createElement('connectable');
				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->{'allowGuestControl'});
				append_if_def($connectable_x, 'connected',         $_->connectable->{'connected'});
				append_if_def($connectable_x, 'startConnected',    $_->connectable->{'startConnected'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$pcnic_x->appendChild($backing_x);
				$pcnic_x->appendChild($connectable_x);
				$pcnic_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($pcnic_x);

				push @{$vm_hw_hash_ref->{'NIC'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualPS2Controller') {
				my $ps2_x = $xml_doc->createElement('VirtualPS2Controller');
				append_if_def($ps2_x, 'key', $_->{'key'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$ps2_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($ps2_x);
			}
			elsif (ref($_) eq 'VirtualPointingDevice') {
				# don't back this up.
			}
			elsif (ref($_) eq 'VirtualSCSIPassthrough') {
				my $pass_x = $xml_doc->createElement('VirtualSCSIPassthrough');
				append_if_def($pass_x, 'controllerKey', $_->{'controllerKey'});
				append_if_def($pass_x, 'key',           $_->{'key'});
				append_if_def($pass_x, 'unitNumber',    $_->{'unitNumber'});

				my $backing_x = $xml_doc->createElement('backing');
				append_if_def($backing_x, 'deviceName',    $_->backing->{'deviceName'});
				append_if_def($backing_x, 'useAutoDetect', $_->backing->{'useAutoDetect'});
				if ($_->{'backing'}->isa('VirtualSCSIPassthroughDeviceBackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualSCSIPassthroughDeviceBackingInfo');
				} else {
					die "Could not identify backing for SCSIPassthrough: " . ref $_->{'backing'} . "\n";
				}

				my $connectable_x = $xml_doc->createElement('connectable');
				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->{'allowGuestControl'});
				append_if_def($connectable_x, 'connected',         $_->connectable->{'connected'});
				append_if_def($connectable_x, 'startConnected',    $_->connectable->{'startConnected'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$pass_x->appendChild($backing_x);
				$pass_x->appendChild($connectable_x);
				$pass_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($pass_x);

				push @{$vm_hw_hash_ref->{'DiskCtrl'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualSerialPort') {
				my $serial_x = $xml_doc->createElement('VirtualSerialPort');

				append_if_def($serial_x, 'controllerKey', $_->controllerKey);
				append_if_def($serial_x, 'key',           $_->key);
				append_if_def($serial_x, 'unitNumber',    $_->unitNumber);
				append_if_def($serial_x, 'yieldOnPoll',   $_->yieldOnPoll);

				my $backing_x = $xml_doc->createElement('backing');

				if ($_->backing->isa('VirtualSerialPortFileBackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualSerialPortFileBackingInfo');
					append_if_def($backing_x, 'fileName', $_->backing->fileName);
				} elsif ($_->backing->isa('VirtualSerialPortDeviceBackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualSerialPortDeviceBackingInfo');
					append_if_def($backing_x, 'deviceName',    $_->backing->deviceName);
					append_if_def($backing_x, 'useAutoDetect', $_->backing->useAutoDetect);
				} elsif ($_->backing->isa('VirtualSerialPortPipeBackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualSerialPortPipeBackingInfo');
					append_if_def($backing_x, 'endpoint', $_->backing->endpoint);
					append_if_def($backing_x, 'noRxLoss', $_->backing->noRxLoss);
					append_if_def($backing_x, 'pipeName', $_->backing->pipeName);
				} elsif ($_->backing->isa('VirtualSerialPortURIBackingInfo')) {
					append_if_def($backing_x, '_backing_type', 'VirtualSerialPortURIBackingInfo');
					append_if_def($backing_x, 'direction',  $_->backing->direction);
					append_if_def($backing_x, 'proxyURI',   $_->backing->proxyURI);
					append_if_def($backing_x, 'serviceURI', $_->backing->serviceURI);
				} else {
					die "Could not identify serial port backing type: " . Dumper($_) . "\n";
				}

				my $connectable_x = $xml_doc->createElement('connectable');
				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->allowGuestControl);
				append_if_def($connectable_x, 'connected',         $_->connectable->connected);
				append_if_def($connectable_x, 'startConnected',    $_->connectable->startConnected);
				append_if_def($connectable_x, 'status',            $_->connectable->status);

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$serial_x->appendChild($backing_x);
				$serial_x->appendChild($connectable_x);
				$serial_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($serial_x);

				push @{$vm_hw_hash_ref->{'Serial'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualSIOController') {
				# This does not necessarily need to be backed up since it's not used during restore.
				my $sio_x = $xml_doc->createElement('VirtualSIOController');
				append_if_def($sio_x, 'key', $_->{'key'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$sio_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($sio_x);
			}
			elsif (ref($_) eq 'VirtualUSBController') {
				# TODO: backing/connectable?
				my $usb_x = $xml_doc->createElement('VirtualUSBController');
				append_if_def($usb_x, 'autoConnectDevices', $_->autoConnectDevices);
				append_if_def($usb_x, 'busNumber',    $_->busNumber);
				append_if_def($usb_x, 'ehciEnabled',  $_->ehciEnabled);
				append_if_def($usb_x, 'key',          $_->key);
				append_if_def($usb_x, 'unitNumber',   $_->unitNumber);

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);  # TODO: retest in 6.2
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$usb_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($usb_x);

				push @{$vm_hw_hash_ref->{'USB'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualUSBXHCIController') {
				my $usb_x = $xml_doc->createElement('VirtualUSBXHCIController');
				append_if_def($usb_x, 'autoConnectDevices', $_->autoConnectDevices);
				append_if_def($usb_x, 'busNumber',          $_->busNumber);
				append_if_def($usb_x, 'key',                $_->key);
				append_if_def($usb_x, 'unitNumber',         $_->unitNumber);

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$usb_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($usb_x);

				push @{$vm_hw_hash_ref->{'USBXHCI'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualUSB') {
				# Don't back this up.  (Attached USB device.)
			}
			elsif (ref($_) eq 'VirtualVmxnet') {
				my $vmxnet_x = $xml_doc->createElement('VirtualVmxnet');
				append_if_def($vmxnet_x, 'addressType',      $_->{'addressType'});
				append_if_def($vmxnet_x, 'controllerKey',    $_->{'controllerKey'});
				append_if_def($vmxnet_x, 'key',              $_->{'key'});
				append_if_def($vmxnet_x, 'macAddress',       $_->{'macAddress'});
				append_if_def($vmxnet_x, 'unitNumber',       $_->{'unitNumber'});
				append_if_def($vmxnet_x, 'wakeOnLanEnabled', $_->{'wakeOnLanEnabled'});

				my $backing_x = $xml_doc->createElement('backing');
				get_nic_backing_type(\$backing_x, \$_->backing);
				append_if_def($backing_x, 'deviceName', $_->backing->{'deviceName'});
				append_if_def($backing_x, 'inPassthroughMode', $_->backing->{'inPassthroughMode'});
				append_if_def($backing_x, 'useAutoDetect',     $_->backing->{'useAutoDetect'});

				my $connectable_x = $xml_doc->createElement('connectable');
				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->{'allowGuestControl'});
				append_if_def($connectable_x, 'connected',         $_->connectable->{'connected'});
				append_if_def($connectable_x, 'startConnected',    $_->connectable->{'startConnected'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$vmxnet_x->appendChild($backing_x);
				$vmxnet_x->appendChild($connectable_x);
				$vmxnet_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($vmxnet_x);

				push @{$vm_hw_hash_ref->{'NIC'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualVmxnet2') {
				my $vmxnet_x = $xml_doc->createElement('VirtualVmxnet2');
				append_if_def($vmxnet_x, 'addressType',      $_->{'addressType'});
				append_if_def($vmxnet_x, 'controllerKey',    $_->{'controllerKey'});
				append_if_def($vmxnet_x, 'key',              $_->{'key'});
				append_if_def($vmxnet_x, 'macAddress',       $_->{'macAddress'});
				append_if_def($vmxnet_x, 'unitNumber',       $_->{'unitNumber'});
				append_if_def($vmxnet_x, 'wakeOnLanEnabled', $_->{'wakeOnLanEnabled'});

				my $backing_x = $xml_doc->createElement('backing');
				get_nic_backing_type(\$backing_x, \$_->backing);
				append_if_def($backing_x, 'deviceName', $_->backing->{'deviceName'});
				append_if_def($backing_x, 'inPassthroughMode', $_->backing->{'inPassthroughMode'});
				append_if_def($backing_x, 'useAutoDetect',     $_->backing->{'useAutoDetect'});

				my $connectable_x = $xml_doc->createElement('connectable');
				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->{'allowGuestControl'});
				append_if_def($connectable_x, 'connected',         $_->connectable->{'connected'});
				append_if_def($connectable_x, 'startConnected',    $_->connectable->{'startConnected'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$vmxnet_x->appendChild($backing_x);
				$vmxnet_x->appendChild($connectable_x);
				$vmxnet_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($vmxnet_x);

				push @{$vm_hw_hash_ref->{'NIC'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualVmxnet3') {
				my $vmxnet_x = $xml_doc->createElement('VirtualVmxnet3');
				append_if_def($vmxnet_x, 'addressType',      $_->{'addressType'});
				append_if_def($vmxnet_x, 'controllerKey',    $_->{'controllerKey'});
				append_if_def($vmxnet_x, 'key',              $_->{'key'});
				append_if_def($vmxnet_x, 'macAddress',       $_->{'macAddress'});
				append_if_def($vmxnet_x, 'unitNumber',       $_->{'unitNumber'});
				append_if_def($vmxnet_x, 'wakeOnLanEnabled', $_->{'wakeOnLanEnabled'});

				my $backing_x = $xml_doc->createElement('backing');
				get_nic_backing_type(\$backing_x, \$_->backing);
				append_if_def($backing_x, 'deviceName',        $_->backing->{'deviceName'});
				append_if_def($backing_x, 'inPassthroughMode', $_->backing->{'inPassthroughMode'});
				append_if_def($backing_x, 'useAutoDetect',     $_->backing->{'useAutoDetect'});

				my $connectable_x = $xml_doc->createElement('connectable');
				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->{'allowGuestControl'});
				append_if_def($connectable_x, 'connected',         $_->connectable->{'connected'});
				append_if_def($connectable_x, 'startConnected',    $_->connectable->{'startConnected'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$vmxnet_x->appendChild($backing_x);
				$vmxnet_x->appendChild($connectable_x);
				$vmxnet_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($vmxnet_x);
				
				push @{$vm_hw_hash_ref->{'NIC'}}, $_->key . ':' . ref($_);
			}
			elsif (ref($_) eq 'VirtualSriovEthernetCard') {
				my $sriov_x = $xml_doc->createElement('VirtualSriovEthernetCard');
				append_if_def($sriov_x, 'addressType',      $_->{'addressType'});
				append_if_def($sriov_x, 'controllerKey',    $_->{'controllerKey'});
				append_if_def($sriov_x, 'key',              $_->{'key'});
				append_if_def($sriov_x, 'macAddress',       $_->{'macAddress'});
				append_if_def($sriov_x, 'unitNumber',       $_->{'unitNumber'});
				append_if_def($sriov_x, 'wakeOnLanEnabled', $_->{'wakeOnLanEnabled'});

				my $backing_x = $xml_doc->createElement('backing');
				get_nic_backing_type(\$backing_x, \$_->backing);
				append_if_def($backing_x, 'deviceName',        $_->backing->{'deviceName'});
				append_if_def($backing_x, 'inPassthroughMode', $_->backing->{'inPassthroughMode'});
				append_if_def($backing_x, 'useAutoDetect',     $_->backing->{'useAutoDetect'});

				my $connectable_x = $xml_doc->createElement('connectable');
				append_if_def($connectable_x, 'allowGuestControl', $_->connectable->{'allowGuestControl'});
				append_if_def($connectable_x, 'connected',         $_->connectable->{'connected'});
				append_if_def($connectable_x, 'startConnected',    $_->connectable->{'startConnected'});

				my $deviceInfo_x = $xml_doc->createElement('deviceInfo');
				append_if_def($deviceInfo_x, 'label',   $_->deviceInfo->label);
				append_if_def($deviceInfo_x, 'summary', $_->deviceInfo->summary);

				$sriov_x->appendChild($backing_x);
				$sriov_x->appendChild($connectable_x);
				$sriov_x->appendChild($deviceInfo_x);
				$device_hw->appendChild($sriov_x);

				push @{$vm_hw_hash_ref->{'NIC'}}, $_->key . ':' . ref($_);
			}
			else {
				die "Unable to identify device of type " . ref($_) . ": $_\n" . Dumper($_) . "\n" .
				"PASSTHROUGHThe VM is configured with one or more unrecognized virtual hardware devices and cannot be backed up with this configuration.".
				"Refer to KB article http://support.unitrends.com/ikm/questions.php?questionid=1500 for steps to resolve this issue.";
			}
		}

		$vm_hw->appendChild($device_hw);

		$vm_x->appendChild($vm_hw);
	}

	$xml_node->appendChild($vm_x);

	print STDERR "Hardware hash: " . Dumper($vm_hw_hash_ref) . "\n";
	my $hw_hash_string = gen_hw_hash_string($vm_hw_hash_ref);
	$xml_node->appendTextChild('vm_hardware_hash', $hw_hash_string);

	return $hw_hash_string;
}

# Returns the version of the current ESX Host to which the VM belongs
sub find_my_esx_version {

	my $vm_uuid = shift;

	my $esx_version;

	my $vm_moref = vim_get_vm_by_uuid($vm_uuid);
	if (not defined $vm_moref) {
		debug_message("Virtual Machine is not present in the inventory. Unable to remove snapshot.");
		goto RETURNVERSION;
	}

	my $vm_view = vim_get_view($vm_moref, "runtime.host");
	if (not defined $vm_view) {
		debug_message("Unable to retrieve properties of the VM for snapshot removal.");
		goto RETURNVERSION;
	}

	my $host_moref = $vm_view->{'runtime.host'};
	if (not defined $host_moref) {
		debug_message("VM's host is not accessible.");
		goto RETURNVERSION;
	}

	my $host_view = vim_get_view($host_moref, "summary.config.product.version");
	if (not defined $host_view) {
		debug_message("Unable to retrieve properties of the ESX host for snapshot removal.");
		goto RETURNVERSION;
	}

	$esx_version = $host_view->{'summary.config.product.version'};
	debug_message("ESX Host's version is $esx_version.");

	RETURNVERSION:
	return $esx_version;

}

sub remove_snapshot {
	# Remove a snapshot we created for the backup.

	# Create a new moref with the identifier we saved previously.
	my $snap_moref_id = shift;
	my $vm_uuid = shift;

	debug_message("Removing snapshot with ID ==> $snap_moref_id");

	my $snap_moref = bless ({ value => $snap_moref_id, type => 'VirtualMachineSnapshot' }, 'ManagedObjectReference');

	my $snap_view = Vim::get_view(
		mo_ref => $snap_moref,
		type => 'VirtualMachineSnapshot',
		properties => [ ] );  # No properties needed, just want to call RemoveSnapshot()

	if (!defined $snap_view) {
		die "Couldn't locate snapshot " . $snap_moref . "\n";
	}

	my $esx_version = find_my_esx_version($vm_uuid);

	my $task_ref;
	eval {
		if ( (not defined $esx_version) or	# If unable to retrieve, then, it's safe without the 'consolidate' option
			($esx_version =~ '^3\.') or	# Although 3.x is no longer supported, the check is there for safety
			($esx_version =~ '^4\.') ) {	# For ESX 4.x, consolidate option is not supported
			debug_message("ESX version is $esx_version. Calling RemoveSnapshot_Task without 'consolidate' option");
			$task_ref = $snap_view->RemoveSnapshot_Task( removeChildren => 0 );
		} else {
			debug_message("ESX version is $esx_version. Calling RemoveSnapshot_Task with 'consolidate' option");
			$task_ref = $snap_view->RemoveSnapshot_Task( removeChildren => 0,
									consolidate => 1 );
		}
	};
	if ($@) {
		if (ref($@) eq 'SoapFault') {
			if (ref($@->detail) eq 'NoPermission') {
				die "Failed to remove snapshot: permission denied -- " . $@->detail->privilegeId . " permission is required in your vCenter role.\n";
			}
			if (ref($@->detail) eq 'FileLocked') {
				die "Failed to remove snapshot: unable to access file " . $@->detail->file . " since it is locked.\n";
			}
			die "SoapFault while removing snapshot: " . $@ . "\n";
		}

		die "Error removing snapshot: " . $@->detail . $@ . "\n";
	}

	my $msg = "Successfully removed snapshot";
	getStatus($task_ref, $msg);
}

sub getStatus {

	debug_message("Entering getStatus");

	my ($taskRef,$message) = @_;

	my $task_view = Vim::get_view(mo_ref => $taskRef);
	# Sometimes it takes a small amount of time to get the view to the 'task'
	sleep 2;
	my $taskinfo = $task_view->info->state->val;
	my $continue = 1;
	while ($continue) {
		my $info = $task_view->info;
		if ($info->state->val eq 'success') {
			debug_message("$message");
			#return $info->result;
			$continue = 0;
		} elsif ($info->state->val eq 'error') {
			my $soap_fault = SoapFault->new;
			$soap_fault->name($info->error->fault);
			$soap_fault->detail($info->error->fault);
			$soap_fault->fault_string($info->error->localizedMessage);
			die "$soap_fault\n";
		} else {
			debug_message("Performing API task...");
			sleep 5;
			$task_view->ViewBase::update_view_data();
		}
	}

	debug_message("Leaving getStatus");
}


#####################################
## Windows Instant Recovery on VMware
#####################################

# Attach the appropriate gustFullName and guestId
sub attach_guest_info {

	my $vm = shift;
	my $os = shift;
	my $cpu = shift;
	my $efi = shift;
	my $xmldoc = shift;

	my $cpu_32 = 0; # True if CPU is 32-bit

	if ( ($cpu eq "Pentium") or ($cpu eq "X86") ) {
		$cpu_32 = 1;
	}

	# Based on the OS the CPU type, attach the appropriate 'guestId' and 'guestFullName'
	if ($os eq "Windows 2003") {
		# For Windows 2003, always create with 64-bit guest OS, since VMware doesn't
		# support the 64-bit "winbm.iso" for 32-bit 2003 VMs
		append_if_def($vm, 'guestFullName', 'Microsoft Windows Server 2003 (64-bit)');
		append_if_def($vm, 'guestId', 'winNetEnterprise64Guest');
	} elsif ($os eq "Windows XP") {
		if ($cpu_32) {
			append_if_def($vm, 'guestFullName', 'Microsoft Windows XP Professional (32-bit)');
			append_if_def($vm, 'guestId', 'winXPProGuest');
		} else {
			append_if_def($vm, 'guestFullName', 'Microsoft Windows XP Professional (64-bit)');
			append_if_def($vm, 'guestId', 'winXPPro64Guest');
		}
	} elsif ($os eq "Windows 2008") {
		if ($cpu_32) {
			append_if_def($vm, 'guestFullName', 'Microsoft Windows Server 2008 (32-bit)');
			append_if_def($vm, 'guestId', 'winLonghornGuest');
		} else {
			append_if_def($vm, 'guestFullName', 'Microsoft Windows Server 2008 (64-bit)');
			append_if_def($vm, 'guestId', 'winLonghorn64Guest');
		}
	} elsif ($os eq "Windows Vista") {
		if ($cpu_32) {
			append_if_def($vm, 'guestFullName', 'Microsoft Windows Vista (32-bit)');
			append_if_def($vm, 'guestId', 'winVistaGuest');
		} else {
			append_if_def($vm, 'guestFullName', 'Microsoft Windows Vista (64-bit)');
			append_if_def($vm, 'guestId', 'winVista64Guest');
		}
	} elsif ($os eq "Windows 2008 R2") {
		# Only 64-bit
		append_if_def($vm, 'guestFullName', 'Microsoft Windows Server 2008 R2 (64-bit)');
		append_if_def($vm, 'guestId', 'windows7Server64Guest');
	} elsif ($os eq "Windows 7") {
		if ($cpu_32) {
			append_if_def($vm, 'guestFullName', 'Microsoft Windows 7 (32-bit)');
			append_if_def($vm, 'guestId', 'windows7Guest');
		} else {
			append_if_def($vm, 'guestFullName', 'Microsoft Windows 7 (64-bit)');
			append_if_def($vm, 'guestId', 'windows7_64Guest');
		}
	} elsif ( ($os eq "Windows 2012") or ($os eq "Windows 2012 R2") ) {
		# Only 64-bit
		# VMware assigns the _same_ guestFullName and guestId for 2012 and 2012 R2
		append_if_def($vm, 'guestFullName', 'Microsoft Windows Server 2012 (64-bit)');
		append_if_def($vm, 'guestId', 'windows8Server64Guest');
	} elsif ( ($os eq "Windows 8") or ($os eq "Windows 8.1")) {
		# VMware assigns the _same_ guestFullName and guestId for 8 and 8.1
		if ($cpu_32) {
			append_if_def($vm, 'guestFullName', 'Microsoft Windows 8 (32-bit)');
			append_if_def($vm, 'guestId', 'windows8Guest');
		} else {
			append_if_def($vm, 'guestFullName', 'Microsoft Windows 8 (64-bit)');
			append_if_def($vm, 'guestId', 'windows8_64Guest');
		}
	} else {
		# We didn't find the guest OS information
		# Depending on the efi flag, create the VM with the appropriate guest OS
		# These choices have been working and tested previously
		if ($efi) {
			append_if_def($vm, 'guestFullName', 'Microsoft Windows Server 2008 R2 (64-bit)');
			append_if_def($vm, 'guestId', 'windows7Server64Guest');
		} else {
			append_if_def($vm, 'guestFullName', 'Microsoft Windows Server 2003 (64-bit)');
			append_if_def($vm, 'guestId', 'winNetEnterprise64Guest');
		}
	}

}

# Determine the number of LSI controllers to attach, depending on the number of disks
sub attach_disk_controller_config {

	my $number_of_disks = shift;
	my $device_hw	    = shift;
	my $efi		    = shift;
	my $os		    = shift;
	my $xmldoc	    = shift;

	my $number_of_controllers = int($number_of_disks/15);

	# 15 disks per controller
	if ( ($number_of_disks % 15) != 0 ) {
		$number_of_controllers = $number_of_controllers + 1;
	}

	print STDERR "Number of Controllers = $number_of_controllers\n";

	my $default_controller = 1;
	if ( ($os eq "Windows 2012 R2") or ($os eq "Windows 8.1") ) {
		# For 2012 R2 and 8.1, DON'T use the default controller
		print STDERR "Guest OS is - $os. Using the non-default controller (VirtualLsiLogicSASController)\n";
		$default_controller = 0;
	}

	my $key = 1000;
	my $unit_number = 3;
	my $bus_number = 0;

	for (my $i=0 ; $i < $number_of_controllers ; $i++) {
		my $lsi;
		if ( ($efi eq 1) or ($default_controller eq 0) ) {
			$lsi = $xmldoc->createElement('VirtualLsiLogicSASController');
		} elsif ($efi eq 0) {
			$lsi = $xmldoc->createElement('VirtualLsiLogicController');
		} else {
			die "Unknown type of BIOS\n";
		}

		append_if_def($lsi, 'busNumber', $bus_number);
		append_if_def($lsi, 'controllerKey', '100');
		append_if_def($lsi, 'hotAddRemove', '1');
		append_if_def($lsi, 'key', $key);
		append_if_def($lsi, 'scsiCtlrUnitNumber', '7');
		append_if_def($lsi, 'unitNumber', $unit_number);

		$device_hw->appendChild($lsi);

		$key = $key + 1;
		$unit_number = $unit_number + 1;
		$bus_number = $bus_number + 1;
	}

}

# Attach configuration for ALL the disks to the XML
sub attach_disk_config {

	my $disk_size_array = shift;
	my $device_hw       = shift;
	my $xmldoc          = shift;

	my @disk_size_ref = @$disk_size_array;
	my $key = 2000;
	my $unit_number = 0;
	my $controller_key = 1000;
	my $counter = 0;	# Keep track of the SCSI node

	foreach my $iter (@disk_size_ref) {

		if ($counter == 7) {
			# Never SCSI x:7
			$key = $key + 1;
			$unit_number = $unit_number + 1;
			$counter = $counter + 1;
		}

		if ($counter == 16) {
			# Move to next LSI Controller
			$counter = 0;
			$controller_key = $controller_key + 1;
			$unit_number = 0;
		}

		my $disk = $xmldoc->createElement('VirtualDisk');

		my $size_in_kb = $iter * 1024 * 1024;

		append_if_def($disk, 'capacityInKB', $size_in_kb);
		append_if_def($disk, 'controllerKey', $controller_key);
		append_if_def($disk, 'key', $key);
		append_if_def($disk, 'unitNumber', $unit_number);

		my $backing = $xmldoc->createElement('backing');
		append_if_def($backing, '_backing_type', 'VirtualDiskFlatVer2BackingInfo');
		append_or_die($backing, 'uuid', '6000C296-f9bd-7a02-0731-2236c76dddf3'); # Placeholder only
		append_if_def($backing, 'diskMode', 'persistent');
		append_if_def($backing, 'thinProvisioned', '1');
		append_if_def($backing, 'writeThrough', '0');

		my $deviceInfo = $xmldoc->createElement('deviceInfo');

		$key = $key + 1;
		$unit_number = $unit_number + 1;

		$disk->appendChild($backing);
		$disk->appendChild($deviceInfo);

		$device_hw->appendChild($disk);

		$counter = $counter + 1;
	}

}

# Similar to get_vm_information_by_moref sub-routine
# Many elements are attached to the XML for completeness only
# Eg: Although a VM's UUID is attached to the XML, it is not assigned to the created VM

sub create_xml_for_vm {

	my $newvmname      = shift;
	my $pool           = shift;
	my $datastore_name = shift;
	my $hostname       = shift;
	my $disksize       = shift;
	my $floppy_path    = shift;
	my $switch         = shift;
	my $efi		   = shift;
	my $os		   = shift;
	my $cpu		   = shift;
	my $nas_ds_name    = shift;
	my $nas_ds_moref   = shift;
	my $boot_disk_num  = shift;
	my $xmldoc_orig    = shift;
	my $xmlroot        = shift;

       #TODO Track H/W Configuration Changes ??? - Create H/W Hash and Keep a track of it ?

	my $xmloutput = XML::LibXML->createDocument;
	$xmloutput->setEncoding("UTF-8");
	$xmlroot = $xmloutput->createElement('vprotect');
	$xmloutput->setDocumentElement($xmlroot);

	my $xmldoc = $xmloutput;

	my $vm = $xmldoc->createElement('virtual_machine');

	append_if_def($vm, 'name', $newvmname);

	#uuid - needed for completeness
	append_if_def($vm, 'uuid', '42360dd1-d7c9-6865-a058-d41dee8f65ba');

	append_if_def($vm, 'version', 'vmx-07');
	append_if_def($vm, 'changeTrackingEnabled', '0');

	attach_guest_info($vm, $os, $cpu, $efi, $xmldoc);

	#guestId, guestFullName, alternateGuestName, annotation - needed for completeness
	if ($efi eq 1) {
		append_if_def($vm, 'firmware', 'efi');
	} elsif ($efi eq 0) {
		append_if_def($vm, 'firmware', 'bios');
	} else {
		die "Unknown type of BIOS. Cannot create a VM.\n";
	}

	append_if_def($vm, 'alternateGuestName', '');
	append_if_def($vm, 'annotation', '');

	append_if_def($vm, 'cpuHotAddEnabled', '0');
	append_if_def($vm, 'cpuHotRemoveEnabled', '0');

	append_if_def($vm, 'maxMksConnections', '40');

	append_if_def($vm, 'memoryHotAddEnabled', '0');

	append_if_def($vm, 'memoryReservationLockedToMax', '0');

	#config
	my $config = $xmldoc->createElement('config');
	{
		my $bootOptions = $xmldoc->createElement('bootOptions');
		append_if_def($bootOptions, 'bootDelay', '0');
		append_if_def($bootOptions, 'enterBIOSSetup', '0');

		my $bootOrder = $xmldoc->createElement('bootOrder');
		#append_if_def($bootOrder, 'testElement', 'Works!');
		$bootOptions->appendChild($bootOrder);

		#extraConfig
			#cpuid.coresPerSocket

		$config->appendChild($bootOptions);
		$vm->appendChild($config);
	}

	my $hardware = $xmldoc->createElement('hardware');

	append_if_def($hardware, 'numCPU', '1');
	append_if_def($hardware, 'memoryMB', '1024');

	my $deviceInfo;
	my $backing;
	my $connectable;

	my $device_hw = $xmldoc->createElement('devices');

		# IDE Controller 1
		my $ide = $xmldoc->createElement('VirtualIDEController');

		append_if_def($ide, 'busNumber', '0');
		append_if_def($ide, 'key', '200');

		$deviceInfo = $xmldoc->createElement('deviceInfo');
		append_if_def($deviceInfo, 'label', 'IDE 0');
		append_if_def($deviceInfo, 'summary', 'IDE 0');

		$ide->appendChild($deviceInfo);
		$device_hw->appendChild($ide);

		# IDE Controller 2
		$ide = $xmldoc->createElement('VirtualIDEController');
		append_if_def($ide, 'busNumber', '1');
		append_if_def($ide, 'key', '201');

		$deviceInfo = $xmldoc->createElement('deviceInfo');
		append_if_def($deviceInfo, 'label', 'IDE 1');
		append_if_def($deviceInfo, 'summary', 'IDE 1');

		$ide->appendChild($deviceInfo);
		$device_hw->appendChild($ide);

		# PS2 Controller
		my $ps2 = $xmldoc->createElement('VirtualPS2Controller');
		append_if_def($ps2, 'key', '300');

		$deviceInfo = $xmldoc->createElement('deviceInfo');
		append_if_def($deviceInfo, 'label',   'PS2 controller 0');
		append_if_def($deviceInfo, 'summary', 'PS2 controller 0');

		$ps2->appendChild($deviceInfo);
		$device_hw->appendChild($ps2);

		# Virtual PCI Controller
		my $pci = $xmldoc->createElement('VirtualPCIController');
		append_if_def($pci, 'busNumber', '0');
		append_if_def($pci, 'key', '100');

		$deviceInfo = $xmldoc->createElement('deviceInfo');
		append_if_def($deviceInfo, 'label', 'PCI controller 0');
		append_if_def($deviceInfo, 'summary', 'PCI controller 0');

		$pci->appendChild($deviceInfo);
		$device_hw->appendChild($pci);

		# Virtual SIO Controller
		my $sio = $xmldoc->createElement('VirtualSIOController');
		append_if_def($sio, 'key', '400');

		$deviceInfo = $xmldoc->createElement('deviceInfo');
		append_if_def($deviceInfo, 'label', 'SIO controller 0');
		append_if_def($deviceInfo, 'summary', 'SIO controller 0');

		$sio->appendChild($deviceInfo);
		$device_hw->appendChild($sio);

		# Virtual Machine VMCI Device
		my $vmci = $xmldoc->createElement('VirtualMachineVMCIDevice');

		append_if_def($vmci, 'allowUnrestrictedCommunication', '0');
		append_if_def($vmci, 'controllerKey', '100');
		append_if_def($vmci, 'key', '12000');
		append_if_def($vmci, 'unitNumber', '17');

		$deviceInfo = $xmldoc->createElement('deviceInfo');
		append_if_def($deviceInfo, 'label', 'VMCI device');
		append_if_def($deviceInfo, 'summary', 'Device on the virtual machine' .
				' PCI bus that provides support for the virtual machine' .
				' communication interface');

		$vmci->appendChild($deviceInfo);
		$device_hw->appendChild($vmci);

		my $disk_size_array = VProtect::Backup::split_keys($disksize);
		my @disk_size_ref = @$disk_size_array;
		my $number_of_disks = scalar(@disk_size_ref);
		print STDERR "Number of disks = " . $number_of_disks . "\n";

		# NOTE - This may change on future ESX versions
		# Current latest ESXi version is 5.5, which supports maximum of 60 disks per VM
		if ($number_of_disks > 60) {
			die "VMware ESXi servers support creation of a virtual machine with maximum 60 disks. " .
			"The original Windows client has " .$number_of_disks. " disks. Hence, Windows Instant Recovery" .
			"of this client on VMware is not possible.\n";
		}

		if ($os eq "Windows 2003") {
			# Attach _only_ the boot disk to IDE 0:0
			# $boot_disk_num=0 indicates first disk in "disk_size_ref" array is the boot disk
			# $boot_disk_num=1 indicates second disk in "disk_size_ref" array is the boot disk
			# .. and so on
			for (my $i = 0 ; $i < $number_of_disks ; $i++) {
				if ($i == $boot_disk_num) {
					# This is the boot disk. Attach it's configuration to the XML
					my $disk = $xmldoc->createElement('VirtualDisk');

					my $size_in_kb = $disk_size_ref[$i] * 1024 * 1024;

					append_if_def($disk, 'capacityInKB', $size_in_kb);
					append_if_def($disk, 'controllerKey', '200');
					append_if_def($disk, 'key', '3000');
					append_if_def($disk, 'unitNumber', '0');

					my $backing = $xmldoc->createElement('backing');
					append_if_def($backing, '_backing_type', 'VirtualDiskFlatVer2BackingInfo');
					append_or_die($backing, 'uuid', '6000C296-f9bd-7a02-0731-2236c76dddf3'); # Placeholder only
					append_if_def($backing, 'diskMode', 'persistent');
					append_if_def($backing, 'thinProvisioned', '1');
					append_if_def($backing, 'writeThrough', '0');

					my $deviceInfo = $xmldoc->createElement('deviceInfo');

					$disk->appendChild($backing);
					$disk->appendChild($deviceInfo);

					$device_hw->appendChild($disk);

					last;
				}
			}

			# Now, remove this element from the disk array,
			# since already attached it to the XML
			splice @$disk_size_array, $boot_disk_num, 1;
		}

		if (scalar(@$disk_size_array) > 0) {
			# Virtual LSI Logic SAS Controller/s
			attach_disk_controller_config($number_of_disks, $device_hw, $efi, $os, $xmldoc);

			# All Virtual Disks
			attach_disk_config($disk_size_array, $device_hw, $xmldoc);
		}

		# Virtual CD-ROM
		my $cdrom = $xmldoc->createElement('VirtualCdrom');

		append_if_def($cdrom, 'controllerKey', '201');
		append_if_def($cdrom, 'key', '3002');
		append_if_def($cdrom, 'unitNumber', '0');

		$deviceInfo = $xmldoc->createElement('deviceInfo');

		append_if_def($deviceInfo, 'label', 'CD/DVD drive 1');
		append_if_def($deviceInfo, 'summary', 'Remote device');

		$connectable = $xmldoc->createElement('connectable');

		append_if_def($connectable, 'allowGuestControl', '1');
		append_if_def($connectable, 'connected', '1');
		append_if_def($connectable, 'startConnected', '1');
		append_if_def($connectable, 'status', 'untried');

		#TODO Add 'backing' information
		$backing = $xmldoc->createElement('backing');

		append_if_def($backing, '_backing_type', 'VirtualCdromIsoBackingInfo');
		# The 'winbm.iso' is always placed on the appliance in under "/usr/bp/wvf/winbm.iso"
		# The below representation is used for VMware to understand the same
		my $cdrom_filename = "[$nas_ds_name] winbm.iso";
		append_if_def($backing, 'fileName', $cdrom_filename);
		#append_if_def($backing, 'useAutoDetect', '0'); # This might be useful later. Change it to TRUE/1i

		$cdrom->appendChild($backing);
		$cdrom->appendChild($connectable);
		$cdrom->appendChild($deviceInfo);
		$device_hw->appendChild($cdrom);

		# Virtual Floppy
		my $floppy = $xmldoc->createElement('VirtualFloppy');

		append_if_def($floppy, 'controllerKey', '400');
		append_if_def($floppy, 'key', '8000');
		append_if_def($floppy, 'unitNumber', '0');

		$deviceInfo = $xmldoc->createElement('deviceInfo');
		append_if_def($deviceInfo, 'label', 'Floppy drive 1');
		append_if_def($deviceInfo, 'summary', 'Remote');

		$connectable = $xmldoc->createElement('connectable');
		append_if_def($connectable, 'allowGuestControl', '1');
		append_if_def($connectable, 'connected', '1');
		append_if_def($connectable, 'startConnected', '1');

		$backing = $xmldoc->createElement('backing');
		#append_if_def($backing, 'deviceName', 'name');
		#append_if_def($backing, 'useAutoDetect', '0');

		append_if_def($backing, '_backing_type', 'VirtualFloppyImageBackingInfo');
		# $main_path will be of the form "<node number of the WIR client>/Static.vfd"
		# Eg: 5/Static.vfd
		my $main_path = substr($floppy_path, 12);
		# The below representation is used for VMware to understand the same
		my $floppy_filename = "[$nas_ds_name] $main_path";
		append_if_def($backing, 'fileName', $floppy_filename);

		$floppy->appendChild($backing);
		$floppy->appendChild($deviceInfo);
		$floppy->appendChild($connectable);
		$device_hw->appendChild($floppy);

		# Virtual NIC
		my $e1000 = $xmldoc->createElement('VirtualE1000');

		append_if_def($e1000, 'addressType', 'assigned');
		append_if_def($e1000, 'controllerKey', '100');
		append_if_def($e1000, 'key', '4000');
		append_if_def($e1000, 'macAddress', '00:50:56:b6:65:e0'); # Placeholder only
		append_if_def($e1000, 'unitNumber', '7');
		append_if_def($e1000, 'wakeOnLanEnabled', '1');

		$backing = $xmldoc->createElement('backing');

		# get_nic_backing_type(\$backing_x, \$_->backing); TODO - Need this ??
		# Might need to attach the switch information later
		append_if_def($backing, '_backing_type', 'VirtualEthernetCardNetworkBackingInfo');
		# Also could be - VirtualEthernetCardDistributedVirtualPortBackingInfo along with switchUuid

		append_if_def($backing, 'deviceName', $switch);
		#append_if_def($backing, 'deviceName', 'VM Network');
		append_if_def($backing, 'inPassthroughMode', '0');
		append_if_def($backing, 'useAutoDetect', '0');

		$connectable = $xmldoc->createElement('connectable');
		append_if_def($connectable, 'allowGuestControl', '1');
		append_if_def($connectable, 'connected', '1');
		append_if_def($connectable, 'startConnected', '1');

		$deviceInfo = $xmldoc->createElement('deviceInfo');
		append_if_def($deviceInfo, 'label', 'Network adapter 1');
		append_if_def($deviceInfo, 'summary', 'VM Network');

		$e1000->appendChild($backing);
		$e1000->appendChild($connectable);
		$e1000->appendChild($deviceInfo);
		$device_hw->appendChild($e1000);

	$hardware->appendChild($device_hw);

	$vm->appendChild($hardware);

	$xmlroot->appendChild($vm);

	return $xmldoc;

} # End create_xml_for_vm

1;
