#!/usr/bin/perl

# Command line options:
#  VMWare standard:
#   --server
#   --username
#   --password (may be removed later)
#   
#  Unitrends:
#   --bpcmd get_vcenter_info
#   --bpcmd get_vm_info (same as above)
#   --bpcmd get_datastore_info
#   --bpcmd get_vm_restore_targets
#   --backup
#   --restore
#   --instant_recovery
#   --removesnap
#   --datastore_name Path to datastore (restore)
#   --hostname Hostname of ESX server (restore)
#   --newvmname New name of VM (restore)
#   --uuid filter by uuid
#   --change_ids
#

use strict;
use warnings;
use 5.008;  # Just to be safe.

use FindBin;
# Find and load VMware's libraries
use lib "$FindBin::RealBin";
use lib "$FindBin::RealBin/vmware/lib";
use lib "$FindBin::RealBin/perllib/VMware/share";
use lib "$FindBin::RealBin/perllib/apps";
use lib "$FindBin::RealBin/../../vmware-vsphere-cli-distrib/apps";
use lib "$FindBin::RealBin/../../vmware-vsphere-cli-distrib/lib/VMware/share/";

package VProtect;

use VProtect 'all';
use Backup 'all';
use Restore 'all';

use Getopt::Long;
use VMware::VIRuntime;
use XML::LibXML;
use Data::Dumper;

my $xmlroot;

my $is_vcenter = 0;

my %opts = (
	# command-line options, parsed by Opts::parse  
	sname => {
		type     => "=s",
		variable => "sname",
		help     => "Snapshot name",
	},
	datacenter => {
		type     => "=s",
		variable => "datacenter",
		help     => "Name of the datacenter",
	},
	pool => {
		type     => "=s",
		variable => "pool",
		help     => "Name of the resource pool",
	},
	host => {
		type     => "=s",
		variable => "host",
		help     => "Name of the host",
	},
	vmname => {
		type     => "=s",
		variable => "vmname",
		help     => "Name of the VM",
	},
	ipaddress => {
		type     => "=s",
		variable => "ipaddress",
		help     => "IP address of the virtual machine",
	},
	guest_os => {
		type     => "=s",
		variable => "guest_os",
		help     => "Guest OS running on the virtual machine",
	},
	uuid => {
		type	=> "=s",
		variable => "uuid",
		help	=> "UUID of ESX server or VM",
	},
	vmuuid => {
		type    => "=s",
		variable => "vmuuid",
		help    => "UUID of VM"
	},
	bpcmd => {
		type   => "=s",
		variable => "bpcmd",
		help   => "Command <get_info, backup>"
	},
	backup => {
		type  => "",
		variable => "backup",
		help => "Perform a backup"
	},
	newvmname => {
		type => "=s",
		variable => "newvmname",
		help => "New name of the VM (restore)"
	},
	datastore_name => {
		type => "=s",
		variable => "datastore_name",
		help => "Name of datastore to restore to"
	},
	hostname => {
		type => "=s",
		variable => "hostname",
		help => "Hostname of ESX server (restore)"
	},
	restore => {
		type => "",
		variable => "restore",
		help => "Perform a restore"
	},
	change_ids => {
		type => "=s",
		variable => "change_ids",
		help => "Change IDs to use for a differential or incremental backup."
	},
	removesnap => {
		type => "=s",
		variable => "snap_moref_id",
		help => "Remove a snapshot by moref id."
	},
	xml_filename => {
		type => "=s",
		variable => "xml_filename",
		help => "Filename to read VM's XML from"
	},
	instant_recovery => {
		type => "",
		variable => "instant_recovery",
		help => "Perform an instant recovery"
	},
	ir_datastore_host => {
		type => "=s",
		variable => "ir_datastore_host",
		help => "Instant recovery datastore host"
	},
	ir_datastore_path => {
		type => "=s",
		variable => "ir_datastore_path",
		help => "Instant recovery datastore path"
	},
	ir_disk_map_path => {
		type => "=s",
		variable => "ir_disk_map_path",
		help => "Instant recovery disk map path"
	},
	ir_audit => {
		type => "",
		variable => "ir_audit",
		help => "Instant recovery auditing mode"
	},
	ir_status_file => {
		type => "=s",
		variable => "ir_status_file",
		help => "vmware_instant_recovery status file"
	},
	remove_ir_datastore => {
		type => "",
		variable => "remove_ir_datastore",
		help => "Remove an instant recovery datastore"
	},
	datastore_name => {
		type => "=s",
		variable => "datastore_name",
		help => "datastore name (for removal)"
	},
	ds_moref_value => {
		type => "=s",
		variable => "ds_moref_value",
		help => "datastore moref value (for removal)"
	},
	ensure_ds_empty => {
		type => "",
		variable => "ensure_ds_empty",
		help => "Ensure datastore is empty"
	},
	ensure_vm_removed => {
		type => "",
		variable => "ensure_vm_removed",
		help => "Ensure VM is removed from the inventory"
	},
	power => {
		type => "=s",
		variable => "power",
		help => "Power ON the VM or not"
	},
	hardware_hash => {
		type => "=s",
		variable => "hardware_hash",
		help => "hardware hash"
	},
	gosusername=> {
		type => "=s",
		variable => "gosusername",
		help => "guest os user name"
	},
	appaware=> {
		type => "=s",
		variable => "appaware",
		help => "guest os app aware flag"
	},
	gospassword=> {
		type => "=s",
		variable => "gospassword",
		help => "guest os pw"
	},
	disklist=> {
		type => "",
		variable => "disklist",
		help => "get list of disks"
	},
	exclude => {
		type => "=s",
		variable => "exclude",
		help => "keys of disks to exclude"
	},
	metadata => {
		type => "=s",
		variable => "metadata",
		help => "restore metadata flag"
	},
	istemplate => {
		type => "=s",
		variable => "istemplate",
		help => "template flag"
	},
	astemplate => {
		type => "=s",
		variable => "astemplate",
		help => "restore as template"
	},
	backupstatus=> {
		type => "=s",
		variable => "backupstatus",
		help => "results of previous backup"
	},
	wir => {
		type => "",
		variable => "wir",
		help => "WIR operation"
	},
	operation => {
		type => "=s",
		variable => "operation",
		help => "Operation to perform"
	},
	numcpu => {
		type => "=s",
		variable => "numcpu",
		help => "Number of CPUs on the Restored VM"
	},
	memory => {
		type => "=s",
		variable => "memory",
		help => "Memory to assign to the Restored VM"
	},
	disksize => {
		type => "=s",
		variable => "disksize",
		help => "Sizes of all disks"
	},
	floppy => {
		type => "=s",
		variable => "floppy",
		help => "floppy name"
	},
	switchname => {
		type => "=s",
		variable => "switchname",
		help => "switch name"
	},
	efi => {
		type => "=s",
		variable => "efi",
		help => "efi or not"
	},
	location => {
		type => "=s",
		variable => "location",
		help => "location of the screenshot"
	},
	os => {
		type => "=s",
		variable => "os",
		help => "windows OS"
	},
	cpu => {
		type => "=s",
		variable => "cpu",
		help => "cpu type"
	},
	delete_vm => {
		type => "",
		variable => "delete_vm",
		help => "delete the VM"
	},
	containertype => {
		type => "=s",
		variable => "containertype",
		help => "type of container"
	},
	containername => {
		type => "=s",
		variable => "containername",
		help => "name of the container"
	},
	bootdisknumber => {
		type => "=s",
		variable => "bootdisknumber",
		help => "boot disk number"
	},
);


###### main exec block

sub main {
	Opts::add_options(%opts);
	Opts::parse();

	if (defined $ENV{'VMPASSWD'}) {
		Opts::set_option('password', $ENV{'VMPASSWD'});
	}
	if (defined $ENV{'VMGOSPASSWD'}) {
		Opts::set_option('gospassword', $ENV{'VMGOSPASSWD'});
	}

	my $xmldoc = init_xml();
	my $warning;

	validate($xmldoc);

	eval {
		Util::connect(); # establish connection to ESX server
	};
	if ($@) {
		chomp $@;
		output_xml_error($xmldoc, "Connection failure: " . $@ . "\n");
		print $xmldoc->toString(1);
		return 1;
	}

	if (not defined $xmldoc) {
		die("xmldoc error");
	}


	if (Opts::option_is_set('backup')) {
		eval {
			VProtect::Backup::do_backup(Opts::get_option('backup'), Opts::get_option('vmuuid'), Opts::get_option('change_ids'), Opts::get_option('hardware_hash'), Opts::get_option('server'), Opts::get_option('appaware'), Opts::get_option('username'),Opts::get_option('password'),  Opts::get_option('gosusername'),Opts::get_option('gospassword'), Opts::get_option('exclude'), Opts::get_option('istemplate'), $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Backup failed: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('restore')) {
		my $datastore_name = Opts::get_option('datastore_name');
		my $hostname = Opts::get_option('hostname');
		my $newvmname = Opts::get_option('newvmname');
		my $metadata = Opts::get_option('metadata');
		my $pool = Opts::get_option('pool');
		my $astemplate = Opts::get_option('astemplate');
		my $xml_filename = Opts::get_option('xml_filename');
		eval {
			VProtect::Restore::do_restore($datastore_name, $hostname, $newvmname, $metadata, $pool, $astemplate, $xml_filename, $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Restore failed: ". $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('instant_recovery')) {
		eval {
			VProtect::Restore::do_instant(
				hostname          => Opts::get_option('hostname'),
				pool		  => Opts::get_option('pool'),
				newvmname         => Opts::get_option('newvmname'),
				datastore_name    => Opts::get_option('datastore_name'),
				ir_datastore_host => Opts::get_option('ir_datastore_host'),
				ir_datastore_path => Opts::get_option('ir_datastore_path'),
				ir_disk_map_path  => Opts::get_option('ir_disk_map_path'),
				xml_filename      => Opts::get_option('xml_filename'),
				ir_audit          => Opts::get_option('ir_audit'),
				ir_status_file    => Opts::get_option('ir_status_file'),
				power		  => Opts::get_option('power'),
				uuid              => Opts::get_option('uuid'),
				# connection info
				server            => Opts::get_option('server'),
				username          => Opts::get_option('username'),
				password          => Opts::get_option('password')
			);
		};
		if ($@) {
			output_xml_error($xmldoc, $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('remove_ir_datastore')) {
		eval {
			if (!Opts::option_is_set('hostname')) {
				die "Hostname argument required.\n";
			}
			if (!Opts::option_is_set('ds_moref_value')) {
				die "ds_moref_value argument required.\n";
			}
			if (!Opts::option_is_set('datastore_name')) {
				die "datastore_name argument required.\n";
			}
			VProtect::Restore::remove_ir_datastore(
				hostname => Opts::get_option('hostname'),
				datastore_moref_value => Opts::get_option('ds_moref_value'),
				datastore_name => Opts::get_option('datastore_name')
			);
		};
		if ($@) {
			output_xml_error($xmldoc, $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('ensure_ds_empty')) {
		eval {
			if (!Opts::option_is_set('hostname')) {
				die "Hostname argument required.\n";
			}
			if (!Opts::option_is_set('ds_moref_value')) {
				die "ds_moref_value argument required.\n";
			}
			VProtect::Restore::ensure_datastore_empty(
				hostname => Opts::get_option('hostname'),
				datastore_moref_value => Opts::get_option('ds_moref_value')
			);
		}; if ($@) {
			output_xml_error($xmldoc, $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('ensure_vm_removed')) {
		eval {
			if (!Opts::option_is_set('vmuuid')) {
				die "VM UUID is required to find it's existence.\n";
			}

			my $vmuuid = Opts::get_option('vmuuid');
			my $ir_ds_name = Opts::get_option('datastore_name');
			my $ir_ds_moref = Opts::get_option('ds_moref_value');

			VProtect::Restore::ensure_vm_removed($vmuuid, $ir_ds_name, $ir_ds_moref);
		}; if ($@) {
			output_xml_error($xmldoc, $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('delete_vm')) {
		eval {
			if (!Opts::option_is_set('vmuuid')) {
				die "VM UUID is required to delete the VM.\n";
			}

			my $vmuuid = Opts::get_option('vmuuid');

			VProtect::Restore::delete_vmir_vm($vmuuid, $xmldoc, $xmlroot);
		}; if ($@) {
			output_xml_error($xmldoc, $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	}
	elsif (Opts::option_is_set('disklist')) {
		eval {
			get_disk_information(Opts::get_option('vmuuid'), Opts::get_option('istemplate'), $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Disk listing failed: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('bpcmd') && Opts::get_option('bpcmd') eq 'get_vm_info') {
		# info
		eval {
			get_vcenter_information($xmldoc);

			get_vsphere_information($xmldoc, 1, 0);
		}; if ($@) {
			if (ref($@) eq 'SoapFault') {
				if (ref($@->detail) eq 'NoPermission') {
					output_xml_error($xmldoc, "Failed to retrieve information: NoPermission: " .
						"Requires permission " . $@->detail->privilegeId . "\n");
					print STDERR Dumper($@);
					print $xmldoc->toString(1);
					Util::disconnect();
					return 1;
				}
			}
			output_xml_error($xmldoc, "Failed to retrieve information: " . $@);
			print STDERR Dumper($@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('bpcmd') && Opts::get_option('bpcmd') eq 'get_datastore_info') {
		eval {
			get_vcenter_information($xmldoc);

			get_vsphere_information($xmldoc, 0, 0);
		}; if ($@) {
			output_xml_error($xmldoc, "Failed to retrieve information: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('bpcmd') && Opts::get_option('bpcmd') eq 'get_network_info') {
		eval {

			get_network_information($xmldoc);

		}; if ($@) {
			output_xml_error($xmldoc, "Failed to retrieve information: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('bpcmd') && Opts::get_option('bpcmd') eq 'get_vm_and_disk_info') {
		eval {
			get_vcenter_information($xmldoc);

			get_vsphere_information($xmldoc, 1, 1);
		};if ($@) {
			output_xml_error($xmldoc, "Failed to retrieve information: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('bpcmd') && ( (Opts::get_option('bpcmd') eq 'get_all_datacenters') or
						(Opts::get_option('bpcmd') eq 'get_all_esx_clusters') or
						(Opts::get_option('bpcmd') eq 'get_all_esx_servers') or
						(Opts::get_option('bpcmd') eq 'get_all_datastores') or
						(Opts::get_option('bpcmd') eq 'get_all_resource_pools') or
						(Opts::get_option('bpcmd') eq 'get_all_vapps') or
						(Opts::get_option('bpcmd') eq 'get_all_folders') ) ) {
		eval {
			my $container = Opts::get_option('bpcmd');
			get_inventory_info($xmldoc, $container);
		}; if ($@) {
			output_xml_error($xmldoc, "Failed to retrieve information: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('removesnap')) {
		my $appflg;
		my $bkstat;
		$appflg = Opts::get_option('appaware');
		print STDERR "uni-vshpere: removename. appaware flag : " . $appflg . " \n";
		if ($appflg eq "YES") {
			$bkstat = Opts::get_option('backupstatus');
			print STDERR "uni-vshpere: removename. backupstatus : " . $bkstat . " \n";
			eval {
			VProtect::Backup::do_backup_conclude(Opts::get_option('uuid'), Opts::get_option('server'), 
				$appflg, Opts::get_option('username'), Opts::get_option('password'), 
				Opts::get_option('gosusername'), Opts::get_option('gospassword'), $bkstat);
			}
		}
		eval {
			remove_snapshot(Opts::get_option('removesnap'), Opts::get_option('uuid'));
		}; if ($@) {
			output_xml_error($xmldoc, "Failed to remove snapshot: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('wir') && Opts::get_option('operation') eq 'createvm'){
		eval {
			print STDERR "Creating a VM for WIR\n";

			my $newvmname      = Opts::get_option('newvmname');
			my $pool           = Opts::get_option('pool');
			my $datastore_name = Opts::get_option('datastore_name');
			my $hostname       = Opts::get_option('hostname');
			my $disksize       = Opts::get_option('disksize');
			my $floppy         = Opts::get_option('floppy');
			my $switch	   = Opts::get_option('switchname');
			my $efi		   = Opts::get_option('efi');
			my $os		   = Opts::get_option('os');
			my $cpu		   = Opts::get_option('cpu');
			my $boot_disk_num  = Opts::get_option('bootdisknumber');

			VProtect::Restore::create_vm_wir($newvmname, $pool, $datastore_name, $hostname, $disksize, $floppy, $switch, $efi, $os, $cpu, $boot_disk_num, $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Failed to Create VM for WIR: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('wir') && Opts::get_option('operation') eq 'restore') {

		eval {
			print STDERR "Restore operations for WIR\n";
			my $vmuuid = Opts::get_option('vmuuid');
			my $numcpu = Opts::get_option('numcpu');
			my $memory = Opts::get_option('memory');
			VProtect::Restore::restore_vm_wir($vmuuid, $numcpu, $memory, $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Failed to perform Restore Mode operation for WIR: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}

	} elsif ( (Opts::option_is_set('wir')) && ( (Opts::get_option('operation') eq 'audit') or (Opts::get_option('operation') eq 'verify') )) {

		eval {
			print STDERR "Audit Mode Operations\n";
			my $vmuuid = Opts::get_option('vmuuid');
			my $numcpu = Opts::get_option('numcpu');
			my $memory = Opts::get_option('memory');
			VProtect::Restore::audit_vm($vmuuid, $numcpu, $memory, $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Failed to perform Audit Mode operation for WIR: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}

	} elsif (Opts::option_is_set('wir') && Opts::get_option('operation') eq 'live') {

		eval {
			print STDERR "Live Mode Operations\n";
			my $vmuuid = Opts::get_option('vmuuid');
			my $numcpu = Opts::get_option('numcpu');
			my $memory = Opts::get_option('memory');
			VProtect::Restore::live_vm($vmuuid, $numcpu, $memory, $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Failed to perform Live Mode operation for WIR: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}

	} elsif (Opts::option_is_set('wir') && Opts::get_option('operation') eq 'poweron') {

		eval {
			print STDERR "Powering ON VM\n";
			my $vmuuid = Opts::get_option('vmuuid');
			my $numcpu = Opts::get_option('numcpu');
			my $memory = Opts::get_option('memory');
			VProtect::Restore::power_on_vm_wir($vmuuid, $numcpu, $memory, $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Failed to Power ON VM for WIR: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}

	} elsif (Opts::option_is_set('wir') && Opts::get_option('operation') eq 'poweroff') {

		eval {
			print STDERR "Powering OFF VM\n";
			my $vmuuid = Opts::get_option('vmuuid');
			VProtect::Restore::power_off_vm_wir($vmuuid, $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Failed to Power OFF VM for WIR: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}

	} elsif (Opts::option_is_set('wir') && Opts::get_option('operation') eq 'ispoweredon') {

		eval {
			print STDERR "Checking if a VM is powered ON\n";
			my $vmuuid = Opts::get_option('vmuuid');
			VProtect::Restore::is_vm_powered_on($vmuuid, $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Failed to check if VM is powered ON: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}

	} elsif (Opts::option_is_set('wir') && Opts::get_option('operation') eq 'ispoweredoff') {

		eval {
			print STDERR "Checking if a VM is powered OFF\n";
			my $vmuuid = Opts::get_option('vmuuid');
			VProtect::Restore::is_vm_powered_off($vmuuid, $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Failed to check if VM is powered OFF: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}

	} elsif (Opts::option_is_set('wir') && Opts::get_option('operation') eq 'delete'){

		eval {
			my $vmuuid = Opts::get_option('vmuuid');
			print STDERR "Deleting VM with UUID $vmuuid\n";
			VProtect::Restore::delete_vm($vmuuid, $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Failed to delete VM that was created for WIR: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} elsif (Opts::option_is_set('wir') && Opts::get_option('operation') eq 'createsnapshot') {

		eval {
			my $vmuuid = Opts::get_option('vmuuid');
			print STDERR "Creating a snapshot for VM with UUID $vmuuid\n";
			VProtect::Restore::create_snapshot_wir($vmuuid, $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Failed to create a snapshot for VM: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}

	} elsif (Opts::option_is_set('wir') && Opts::get_option('operation') eq 'revertsnapshot') {

		eval {
			my $vmuuid = Opts::get_option('vmuuid');
			print STDERR "Reverting to snapshot for VM with UUID $vmuuid\n";
			VProtect::Restore::revert_snapshot_wir($vmuuid, $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Failed to perform revert snapshot operation: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}

	} elsif (Opts::option_is_set('wir') && Opts::get_option('operation') eq 'removesnapshot') {

		eval {
			my $vmuuid = Opts::get_option('vmuuid');
			print STDERR "Removing snapshot for VM with UUID $vmuuid\n";
			VProtect::Restore::remove_snapshot_wir($vmuuid, $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Failed to perform remove snapshot operation: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}

	} elsif (Opts::option_is_set('wir') && Opts::get_option('operation') eq 'takescreenshot') {

		eval {
			my $vmuuid = Opts::get_option('vmuuid');
			my $location = Opts::get_option('location');
			print STDERR "Taking screenshot for VM with UUID $vmuuid\n";
			VProtect::Restore::take_screenshot_wir($vmuuid, $location, $xmldoc, $xmlroot);
		};
		if ($@) {
			output_xml_error($xmldoc, "Failed to perform create snapshot operation: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}

	} elsif (Opts::option_is_set('containertype')) {
		eval {
			my $container_type = Opts::get_option('containertype');
			my $container_name = Opts::get_option('containername');
			get_container_vms($xmldoc, $container_type, $container_name);
		};
		if ($@) {
			output_xml_error($xmldoc, "Failed to get VMs for this container: " . $@);
			print $xmldoc->toString(1);
			Util::disconnect();
			return 1;
		}
	} else {
		output_xml_error($xmldoc, "Invalid arguments.");
		print $xmldoc->toString(1);
		Util::disconnect();
		exit(2);
	}

	print $xmldoc->toString(1);

	Util::disconnect();

	return 0;
}

###### end main exec

###### functions

sub init_xml () {
	# create a new XML Document
	my $xmloutput = XML::LibXML->createDocument;
	#$xmloutput->setEncoding("UTF-8");

	# create root of the XML Document
	$xmlroot = $xmloutput->createElement('vprotect');
	$xmloutput->setDocumentElement($xmlroot);

	return $xmloutput;
}

sub append_to_xml ($$$) {
	my $xml_elem = shift;
	my $field = shift;
	my $text = shift;

	if (defined $text) {
		$xml_elem->appendTextChild($field, $text);
	}
}

sub xml_get_or_create ($$) {
	my $xml_target = shift;
	my $name = shift;

	my @list = $xml_target->getElementsByTagName($name);

	if (scalar @list == 0) {
		return $xml_target->createElement($name);
	}

	return $list[0];
}

sub get_vcenter_information ($) {
	my $xml_target = shift;

	debug_message('beginning of get_vcenter_information');

	my $content = vim_get_service_content();
	my $about = $content->about;

	my $vpinfo_node = xml_get_or_create($xml_target, 'vprotect_info');

	print STDERR "apiType is - " .$about->{'apiType'}. "\n";

	if ($about->{'apiType'} ne 'VirtualCenter') {
		# not a vcenter server, bail 
		$xmlroot->appendChild($vpinfo_node);
		return;
	}

	if ($about->{'version'} =~ "^2\.5\.") {
		die "VirtualCenter 2.5 is not supported.\n";
	}

	$is_vcenter = 1;

	my $server_xmlelem = $xml_target->createElement('vcenter_server');

	append_to_xml($server_xmlelem, "apiVersion", $about->{'apiVersion'});
	append_or_die($server_xmlelem, "instanceUuid", $about->{'instanceUuid'});
	append_or_die($server_xmlelem, "uuid", $about->{'instanceUuid'});
	append_to_xml($server_xmlelem, "fullName", $about->{'fullName'});
	append_to_xml($server_xmlelem, "licenseProductName", $about->{'licenseProductName'});
	append_to_xml($server_xmlelem, "licenseProductVersion", $about->{'licenseProductVersion'});
	append_to_xml($server_xmlelem, "osType", $about->{'osType'});
	append_to_xml($server_xmlelem, "version", $about->{'version'});

	my $vpx_settings = $content->setting;
	debug_message('get_view vpx_settings');
	my $settings_view = Vim::get_view(
			mo_ref => $vpx_settings,
			properties => [ ]  # No properties, just want to get access to QueryOptions()
	);
	debug_message('end get_view vpx_settings');
	my $result = $settings_view->QueryOptions(name => 'VirtualCenter.InstanceName');
	my $name = @$result[0]->value;

	append_to_xml($server_xmlelem, "name", $name);

	$vpinfo_node->appendChild($server_xmlelem);

	$xmlroot->appendChild($vpinfo_node);

	debug_message('end of get_vcenter_information');
}

sub get_vsphere_information ($$) {
	my $xml_target = shift;
	my $get_vm_info = shift;
	my $get_vm_and_disk_info = shift;

	# [ $get_vm_info, $get_vm_and_disk_info ]
	# [ 1, 0 ] --- Return information only for VMs
	# [ 1, 1 ] --- Return information for VMs and the disks for each VM
	# [ 0, 0 ] --- Return information for Datastores

	debug_message('beginning of get_vsphere_information');

	# Check for ESXi free
	my $content = vim_get_service_content();
	my $lm = $content->licenseManager;
	debug_message('get_view lm');
	my $lm_view = Vim::get_view( mo_ref => $lm );
	debug_message('end get_view lm');

	my $esxi_free = 0;
	if ($content->about->productLineId eq 'embeddedEsx' and $lm_view->licensedEdition eq 'esxBasic') {
		$esxi_free = 1;
	}
	if ($content->about->productLineId eq 'embeddedEsx') {
		if (defined $lm_view->licenses) {
			if (defined $lm_view->licenses->[0]) {
				if ( ($lm_view->licenses->[0]->editionKey =~ /esxBasic/i) ) {
					$esxi_free = 1;
				}
			}
		}
	}
	if ($esxi_free) {
		# This is an ESXi free server, do not return information for it.
		output_xml_error($xml_target, "ESXi free is not supported.");
		return;
	}

	my $esx_name;
	my $vpinfo_xmlelem = xml_get_or_create($xml_target, 'vprotect_info');

	if ($get_vm_info) {
		# Attach all VMs, RPs and vApps for Backup listing

		# This call will be time consuming based on the number of ESX servers in the inventory
		# More number of ESX servers = more time taken for this API to return
		debug_message('vim_get_hosts');
		my $esx_server_views = Vim::find_entity_views(view_type => "HostSystem",
								properties => [ 'name',
										'summary',
										#'config',
										'hardware.cpuInfo',
										'hardware.cpuInfo.numCpuPackages',
										'parent' ] );
		debug_message('end vim_get_hosts');
	
		debug_message('vim_get_compute_resource');
		my $cr_views =  Vim::find_entity_views(view_type => "ComputeResource",
						properties => [ 'name' ] );
		debug_message('end vim_get_compute_resource');

		debug_message('vim_get_resource_pool');
		my $rp_views = Vim::find_entity_views(view_type => "ResourcePool",
						properties => [ 'name',
								'parent',
								'summary' ] );
		debug_message('end vim_get_resource_pool');

		# This call will be time consuming based on the number of VMs in the inventory
		# More number of VMs = more time taken for this API to return
		debug_message('vim_get_virtual_machine');
		my $vm_views = Vim::find_entity_views(view_type => "VirtualMachine",
							properties => [ 'runtime.host',
									'summary',
									'datastore',
									'capability.changeTrackingSupported',
									'config',
									'resourcePool' ] );
		debug_message('end vim_get_virtual_machine');

		print STDERR "Total number of ESX servers              - " .scalar(@{$esx_server_views}). "\n";
		print STDERR "Total number of Virtual Machines         - " .scalar(@{$vm_views}). "\n";
		print STDERR "Total number of Resource Pools and vApps - " .scalar(@{$rp_views}). "\n";

		my %esx_cluster_hash;		# Store ALL the cluster's Mo-Refs here
		my %rp_hash;			# Store ALL the RPs Mo-Refs and their parent's Mo-Refs here
		my %esx_server_hash;		# Store ALL the ESX servers Mo-Refs and their parent's Mo-Refs here

		# Look for all the ESX clusters present in the inventory
		# "esx_cluster_hash" is of the form/pair -- <ESX Cluster Mo-Ref> <===> <0>
		foreach ( @{$cr_views || []} ) {
			my $cr_type = $_->get_property('mo_ref.type');
			if ($cr_type eq "ClusterComputeResource") {
				# Push a 'cluster' element's moref only
				my $cr_value = $_->get_property('mo_ref.value');
				$esx_cluster_hash{$cr_value} = 0;
			}
		}
		if (%esx_cluster_hash) {
			print STDERR "ESX Cluster hash is -\n" .Dumper(%esx_cluster_hash). "\n";
		} else {
			print STDERR "There are no ESX clusters in the inventory\n";
		}

		# Create a hash of "<RP moref> <===> <RP's parent moref>"
		# Helps when identifying if a particular RP/vApp belongs to the current ESX server
		foreach ( @{$rp_views || []} ) {
			$rp_hash{$_->get_property('mo_ref.value')} = $_->get_property('parent')->value;
		}
		print STDERR "RP hash is -\n" . Dumper(%rp_hash) ."\n";

		# Create a hash of "<ESX moref> <===> <ESX's parent moref>"
		foreach ( @{$esx_server_views || []} ) {
			$esx_server_hash{$_->get_property('mo_ref.value')} = $_->get_property('parent')->value;
		}
		print STDERR "ESX server hash is -\n" .Dumper(%esx_server_hash). "\n";

		# All the helper information for creating the XML is ready.
		# Now attach the VMs, RPs and vApps information for all the ESX servers
	
		my $vpinfo_xmlelem = xml_get_or_create($xml_target, 'vprotect_info');

		debug_message('Starting to traverse the inventory');

		# Main loop to collect all ESX servers, VMs, RPs and vApps information
		foreach ( @{$esx_server_views || []} ) {
			print STDERR "Looking at ESX server - " . $_->get_property('name') . "\n";

			my $esx_available = check_esx_availability($_);
			if ($esx_available) {
				print STDERR "This ESX is accessible. Returning information for VMs, " .
				"RPs and vApps for this ESX server\n";
			} else {
				print STDERR "This ESX is currently not accessible. Skipping it.\n";
				next;
			}
			
			# This ESX hist is reachable. Good. Collect all information for it.
			# Create an esx_server XML element for the information.
			my $server_xmlelem = $xml_target->createElement('esx_server');

			attach_esx_information_for_vms($xml_target, $server_xmlelem, $_);

			# Indicates whether to attach only the root RP "Resources"
			# information or all the RPs/VAs information
			my $get_all_rp_va_info = 0;
		
			my $current_esx_moref = $_->get_property('mo_ref.value');
			my $current_esx_parent_moref = $_->get_property('parent')->value;
			my $current_esx_parent_type = $_->get_property('parent')->type;
			print STDERR "Current ESX Server is - " . $_->get_property('summary.config.name') . "\n";

			if (%esx_cluster_hash) {
				my $key;	my $value;
				if ($current_esx_parent_type eq "ClusterComputeResource") {
					# ESX server belongs to a Cluster
					while (($key, $value) = each (%esx_cluster_hash)) {
						if ($key eq $current_esx_parent_moref) {
							if ($value == 0) {
								$esx_cluster_hash{$key} = 1;
								$get_all_rp_va_info = 1;
							}
						}
					}
					if ($get_all_rp_va_info) {
						print STDERR "Changed cluster hash is -\n" .Dumper(%esx_cluster_hash). "\n";
					}
				} else {
					# This ESX server doesn't belong to any ESX Cluster
					# Hence, get all the information about it's RPs and vApps
					$get_all_rp_va_info = 1;
				}
			} else {
				# There are no ESX clusters in the inventory.
				# Get all the information about this ESX server's RPs and vApps
				$get_all_rp_va_info = 1;
			}

			$esx_name = $_->get_property('summary.config.name');
			print STDERR "Attaching VM, RP and vApp information for ESX - $esx_name\n";

			my $current_esx_uuid = $_->get_property('summary.hardware.uuid');
			debug_message('Getting VM information');
			get_vm_information($xml_target, $server_xmlelem, $current_esx_uuid, $current_esx_moref, $current_esx_parent_moref, $esx_server_views, $vm_views, \%rp_hash, $get_vm_and_disk_info);
			debug_message('Got VM information');

			if ($get_all_rp_va_info) {
				# Attach all the RP related information for this ESX server
				debug_message('Attaching all information for Resource Pools and vApps for this ESX server');
				get_pool_information($xml_target, $server_xmlelem, $current_esx_moref, $current_esx_parent_moref, $rp_views, \%rp_hash);
				get_va_information($xml_target, $server_xmlelem, $current_esx_moref, $current_esx_parent_moref, $rp_views, \%rp_hash);
				debug_message('Attached RP and VA info');
			} else {
				debug_message('Attaching only the root Resources for this ESX server');
				get_only_root_pool_information($xml_target, $server_xmlelem, $current_esx_parent_moref, $rp_views);
				debug_message('Attached root Resources info');
			}

			$vpinfo_xmlelem->appendChild($server_xmlelem);

		}	# End foreach esx_server_views

		debug_message('Done traversing the inventory');

	} else {
		# Attach datastore information (get_datastore_info) for Restore side ($get_vm_info = 0)

		# This call will be time consuming based on the number of ESX servers in the inventory
		# More number of ESX servers = more time taken for this API to return
		debug_message('vim_get_hosts');
		my $esx_server_views = Vim::find_entity_views(view_type => "HostSystem",
								properties => [ 'name',
										'summary',
										#'config',
										'datastore' ] );
		debug_message('end vim_get_hosts');

		debug_message('vim_get_datastores');
		my $datastore_views = Vim::find_entity_views(view_type => "Datastore",
							properties => [ 'name',
									'summary' ] );
		debug_message('vim_get_datastores');

		print STDERR "Total number of ESX servers - " .scalar(@{$esx_server_views}). "\n";
		print STDERR "Total number of datastores  - " .scalar(@{$datastore_views}). "\n";

		foreach ( @{$esx_server_views || []} ) {
			print STDERR "Looking at ESX server - " . $_->get_property('name') . "\n";

			my $esx_available = check_esx_availability($_);
			if ($esx_available) {
				print STDERR "This ESX is accessible. Returning information for Datastores, " .
				"for this ESX server\n";
			} else {
				print STDERR "This ESX is currently not accessible. Skipping it.\n";
				next;
			}

			my $server_xmlelem = $xml_target->createElement('esx_server');
			attach_esx_information_for_datastores($xml_target, $server_xmlelem, $_);

			$esx_name = $_->get_property('summary.config.name');
			print STDERR "Attaching datastore information for ESX - $esx_name\n";

			my %datastore_map = map { $_->get_property('mo_ref.value') => $_ } @{ $datastore_views || [] };
			my @datastore_references = map($_->{'value'}, @{$_->get_property('datastore') || []});
			my @datastores = @datastore_map{@datastore_references};

			attach_datastore_information($xml_target, $server_xmlelem, \@datastores);

			$vpinfo_xmlelem->appendChild($server_xmlelem);
		}

	}

	$xmlroot->appendChild($vpinfo_xmlelem);
	debug_message('end of get_vsphere_information');
}


sub check_esx_availability {

	my $current_esx = shift;

	if ( !(defined $_->get_property('summary')) ||
		!(defined $_->get_property('summary.config')) ) {
		print STDERR "Warning: unreachable host (summary not defined)\n";
		return 0;
	}

	#unless (defined $_->get_property('config')) {
	#	print STDERR "Warning: unreachable host (config not defined)\n";
	#	return 0;
	#}

	my $accessibility = $_->get_property('summary.runtime.connectionState')->val;
	debug_message("Connection state of current ESX host - $accessibility");
	unless ($accessibility eq "connected") {
		print STDERR "This ESX is not accessible currently. Skipping it.\n";
		return 0;
	}

	# Can retrieve information for this ESX. Good.
	return 1;

}

# Sub-routine to attach each ESX server's information to the XML (for Backups/Listing of VMs)
sub attach_esx_information_for_vms {

	my $xml_target = shift;
	my $server_xmlelem = shift;
	my $current_esx = shift;

	$server_xmlelem->appendTextChild('name', $current_esx->get_property('summary.config.name'));

	append_to_xml($server_xmlelem, 'managementServerIp', $current_esx->get_property('summary.managementServerIp'));

	if (defined $current_esx->get_property('summary.config.product')) {
		append_to_xml($server_xmlelem, 'fullName', $current_esx->get_property('summary.config.product.fullName'));
		append_to_xml($server_xmlelem, 'apiVersion', $current_esx->get_property('summary.config.product.apiVersion'));
		append_to_xml($server_xmlelem, 'licenseProductName', $current_esx->get_property('summary.config.product.licenseProductName'));
		append_to_xml($server_xmlelem, 'licenseProductVersion', $current_esx->get_property('summary.config.product.licenseProductVersion'));
		append_to_xml($server_xmlelem, 'version', $current_esx->get_property('summary.config.product.version'));
		append_to_xml($server_xmlelem, 'productLineId', $current_esx->get_property('summary.config.product.productLineId'));
		append_to_xml($server_xmlelem, 'build', $current_esx->get_property('summary.config.product.build'));
		append_to_xml($server_xmlelem, 'name', $current_esx->get_property('summary.config.product.name'));
	}

	# Use UUID rather than instanceUUID.
	my $current_esx_uuid = $current_esx->get_property('summary.hardware.uuid');
	append_to_xml($server_xmlelem, 'uuid', $current_esx_uuid);

	##################
	# CPU information
	##################
	if (defined $current_esx->get_property('hardware.cpuInfo')) {
		append_to_xml($server_xmlelem, 'numCpuPackages', $current_esx->get_property('hardware.cpuInfo.numCpuPackages'));
	} else {
		print STDERR "Unable to read hardware->cpuInfo.\n";
	}

}   # End 'attach_esx_information_for_vms' sub-routine


# Sub-routine to attach each ESX server's information to the XML (for Restore side)
sub attach_esx_information_for_datastores {

	my $xml_target = shift;
	my $server_xmlelem = shift;
	my $current_esx = shift;

	$server_xmlelem->appendTextChild('name', $current_esx->get_property('summary.config.name'));

	append_to_xml($server_xmlelem, 'version', $current_esx->get_property('summary.config.product.version'));

	# Use UUID rather than instanceUUID.
	my $current_esx_uuid = $current_esx->get_property('summary.hardware.uuid');
	append_to_xml($server_xmlelem, 'uuid', $current_esx_uuid);

}   # End 'attach_esx_information_for_datastores' sub-routine


# This retrieves information for each virtual machine belonging to the ESX server with UUID "$host_uuid"
sub get_vm_information {

	my $xml_target = shift;
	my $server_xmlelem = shift;
	my $current_esx_uuid = shift;
	my $current_esx_moref = shift;
	my $current_esx_parent_moref = shift;
	my $esx_server_views = shift;
	my $vm_views = shift;
	my $rp_hash = shift;
	my $get_vm_and_disk_info = shift;

	debug_message('begin get_vm_information');

	my $template_parent_key;

	my %esx_server_map = map { $_->get_property('mo_ref.value') => $_ } @{ $esx_server_views || [] };

	my $vms_xmlelem = $xml_target->createElement('virtual_machines');

	foreach ( @{$vm_views || []} ) {
		my $vm_created_for_ir = 0;
		my $vm_name;
		# Get HostSystem from the host map
		my $host_ref = $_->get_property('runtime.host')->{'value'};
		if ( (not defined $host_ref) ) {
			print STDERR "Can't find host ref\n";
			next;
		}
		my $host = $esx_server_map{$host_ref};

		if (not defined $host) {
			print STDERR "Unable to access host information for VM - " .
				$_->summary->config->name. ". Skipping it.\n";
			next;
		}

		if ( (not defined $host->get_property('summary.hardware.uuid') ) ) {
			print STDERR "Cant find host UUID\n";
			next;
		}

		if ($current_esx_uuid eq $host->get_property('summary.hardware.uuid')) {
			# This VM belongs to this ESX server
			my $accessibility = $_->get_property('summary.runtime.connectionState')->val;
			unless ($accessibility eq "connected") {
				# Skip VMs that aren't accessible.
				$vm_name = $_->get_property('summary.config.name');
				print STDERR "VM $vm_name is not accessible or connected currently. Skipping it.\n";
				next;
			}

			if (not defined($_->get_property('datastore'))) {
				# Skip VMs that aren't accessible.
				print STDERR "Skipping VM, datastore not accessible.\n";
				next;
			}

			if (defined $_->get_property('config.ftInfo')) {
				# Skip VMs with fault tolerance enabled.
				print STDERR "Skipping VM " . $_->get_property('config.name') . ", fault tolerance enabled.\n";
				next;
			}

			if (defined $_->config->extraConfig) {
				my $extra_config = $_->config->extraConfig;
				foreach my $entry (@$extra_config) {
					if ($entry->key eq 'unitrends.vm.type') {
						if($entry->value eq 'instant_recovery_audit') {
							print STDERR "Skipping VM " . $_->config->name . " since" .
							" it was created for VMware IR\n";
							$vm_created_for_ir = 1;
							last;
						}
					}
				}
			}
			if($vm_created_for_ir) {
				next;
			}

			if ( (not defined $_->resourcePool) and ($_->config->template eq 0) ) {
				print STDERR "Skipping VM " .$_->get_property('config.name'). ", Resource Pool not accessible";
				next;
			}

			my $vm_xmlelem = $xml_target->createElement('virtual_machine');

			if (defined $_->get_property('config.name')) {
				$vm_name = $_->get_property('config.name');
				$vm_name = translate_name($vm_name, 1);
				$vm_xmlelem->appendTextChild('name', $vm_name);
			}

			if (not defined ($_->get_property('config.uuid')) ) {
				print STDERR "UUID not found. Skip\n";
				next;
			}

			if (not defined ($_->get_property('config.instanceUuid')) ) {
				print STDERR "Instance UUID not found. Skip\n";
				next;
			}

			my $vm_uuid = $_->get_property('config.uuid');
			if (defined $vm_uuid) {
				$vm_xmlelem->appendTextChild('uuid', $vm_uuid);
			}

			my $vm_instance_uuid = $_->get_property('config.instanceUuid');
			if (defined $vm_instance_uuid) {
				 $vm_xmlelem->appendTextChild('instanceUuid', $vm_instance_uuid);
			}

			if (defined($_->get_property('config.version'))) {
				$vm_xmlelem->appendTextChild('version', $_->get_property('config.version'));
			}

			if (defined($_->get_property('config.changeTrackingEnabled'))) {
				$vm_xmlelem->appendTextChild('changeTrackingEnabled', $_->get_property('config.changeTrackingEnabled'));
			}

			if (defined($_->get_property('capability.changeTrackingSupported'))) {
				my $ctk_supported = $_->get_property('capability.changeTrackingSupported');
				if ($ctk_supported eq "true") {
					$vm_xmlelem->appendTextChild('changeTrackingSupported', '1');
				} else {
					$vm_xmlelem->appendTextChild('changeTrackingSupported', '0');
				}
			}

			if ($_->get_property('config.template') eq 0) {
				# This is a VM
				$vm_xmlelem->appendTextChild('parentKey', $_->get_property('resourcePool')->value);
				my $parent_type_found = 0;
				my $key;	my $value;
				my $current_rp = $_->get_property('resourcePool')->value;
				while (($key, $value) = each(%$rp_hash)) {
					if ($key eq $current_rp) {
						if ($value eq $current_esx_parent_moref) {
							# VM belongs to ESX server
							$vm_xmlelem->appendTextChild('parentType', 'ComputeResource');
							$parent_type_found = 1;
						}
					}
				}
				if (!$parent_type_found) {
					$vm_xmlelem->appendTextChild('parentType', $_->get_property('resourcePool')->type);
				}
			} else {
				# This is a Template
				$vm_xmlelem->appendTextChild('parentKey', $current_esx_moref);
				$vm_xmlelem->appendTextChild('parentType', 'ComputeResource');
			}

			if (defined($_->get_property('config.template'))) {
				$vm_xmlelem->appendTextChild('template', $_->get_property('config.template'));
			}

			if (defined($_->get_property('config.guestFullName'))) {
				$vm_xmlelem->appendTextChild('guestFullName', $_->get_property('config.guestFullName'));
			}

			if ($get_vm_and_disk_info) {
				my $devices = $_->config->hardware->device;
				attach_disk_info($devices, $vm_name, $vm_xmlelem, $xml_target);
			}

			$vms_xmlelem->appendChild($vm_xmlelem);
		} # Attached all information for this VM
		
	} # End foreach vm_views. Retrieved information for all the VMs on this ESX server

	$server_xmlelem->appendChild($vms_xmlelem);

	debug_message('end get_vm_information');

}   # End 'get_vm_information' sub-routine


# This retrieves information for all Resource Pools belonging to the ESX server with "$current_esx_moref"
sub get_pool_information {

	my $xml_target = shift;
	my $server_xmlelem = shift;
	my $current_esx_moref = shift;
	my $current_esx_parent_moref = shift;
	my $rp_views = shift;
	my $rp_hash = shift;

	debug_message('begin get_pool_information');

	# Now get Resource Pool information
	my $rps_xmlelem = $xml_target->createElement('resource_pools');

	foreach ( @{$rp_views || []} ) {
		# vApps have an 'instanceUuid', RPs don't
		if (not defined $_->get_property('summary.instanceUuid')) {
			# Find if this RP belongs to this ESX server
			print STDERR "Looking at Resource Pool - " .$_->get_property('name'). "\n";
			my $parent_found = 0;
			my $current_rp_moref = $_->get_property('mo_ref.value');

			find_my_parent($rp_hash, $current_rp_moref, $current_esx_moref, $current_esx_parent_moref, \$parent_found);

			print STDERR "Value of parent_found is $parent_found\n";
			if ($parent_found) {
				# Attach RP information only if this RP belongs to this ESX server
				my $rp_name = $_->get_property('name');
				$rp_name = translate_name($rp_name, 1);
				print STDERR "Attaching RP information for -- " .$rp_name. "\n";

				my $rp_xmlelem = $xml_target->createElement('resource_pool');
				$rp_xmlelem->appendTextChild('key', $current_rp_moref);
				$rp_xmlelem->appendTextChild('name', $rp_name);
				$rp_xmlelem->appendTextChild('parentKey', $_->get_property('parent')->value);

				if ($_->get_property('parent')->value eq $current_esx_parent_moref) {
					# Root RP called 'Resources'
					$rp_xmlelem->appendTextChild('parentType', "ComputeResourceDefault");
				} else {
					# If the RP belongs directly to the ESX server, then, need to attach the parentType as "ComputeResource"
					my $key;        my $value;
					my $parent_is_esx = 0;
					my $parent;
					while ( ($key, $value) = each (%$rp_hash) ) {
						if ($key eq $current_rp_moref) {
							$parent = $value;
						}
					}

					while ( ($key, $value) = each (%$rp_hash) ) {
						if ($key eq $parent) {
							if ($value eq $current_esx_parent_moref) {
								$parent_is_esx = 1;
								last;
							}
						}
					}

					if ($parent_is_esx) {
						$rp_xmlelem->appendTextChild('parentType', "ComputeResource");
					} else {
						$rp_xmlelem->appendTextChild('parentType', $_->get_property('parent')->type);
					}
				}

				$rps_xmlelem->appendChild($rp_xmlelem);
			}
		}

	}	# End foreach rp_views

	$server_xmlelem->appendChild($rps_xmlelem);

	debug_message('end get_pool_information');

} # End 'get_pool_information' sub-routine


# This retrieves information for all vApps belonging to the ESX server with "$current_esx_moref"
sub get_va_information {

	my $xml_target = shift;
	my $server_xmlelem = shift;
	my $current_esx_moref = shift;
	my $current_esx_parent_moref = shift;
	my $rp_views = shift;
	my $rp_hash = shift;

	debug_message('begin get_va_information');

	my $vapp_xmlelem;

	# Now get Resource Pool information
	my $vapps_xmlelem = $xml_target->createElement('virtual_apps');

	foreach ( @{$rp_views || []} ) {
		if ($_->get_property('parent')->value eq $current_esx_parent_moref) {
			# Root RP. Need to attach this, since required by core for DB operations
			$vapp_xmlelem = $xml_target->createElement('virtual_app');
			$vapp_xmlelem->appendTextChild('key', $_->get_property('mo_ref.value'));
			$vapp_xmlelem->appendTextChild('name', $_->get_property('name'));
			$vapp_xmlelem->appendTextChild('parentKey', $_->get_property('parent')->value);
			$vapp_xmlelem->appendTextChild('parentType', "ComputeResourceDefault");
			$vapps_xmlelem->appendChild($vapp_xmlelem);
			next;
		}

		# vApps have an 'instanceUuid', RPs don't
		if (defined $_->get_property('summary.instanceUuid')) {
			# Find if this vApp belongs to this ESX server
			print STDERR "Looking at vApp - " .$_->get_property('name'). "\n";
			my $parent_found = 0;
			my $current_vapp_moref = $_->get_property('mo_ref.value');

			find_my_parent($rp_hash, $current_vapp_moref, $current_esx_moref, $current_esx_parent_moref, \$parent_found);

			print STDERR "Value of parent_found is $parent_found\n";
			if ($parent_found) {
				# Attach vApp information only if this vApp belongs to this ESX server
				my $va_name = $_->get_property('name');
				$va_name = translate_name($va_name, 1);
				print STDERR "Attaching vApp information for - " .$va_name. "\n";

				$vapp_xmlelem = $xml_target->createElement('virtual_app');
				$vapp_xmlelem->appendTextChild('key', $current_vapp_moref);
				$vapp_xmlelem->appendTextChild('name', $va_name);
				$vapp_xmlelem->appendTextChild('parentKey', $_->get_property('parent')->value);

				# If the vApp belongs directly to the ESX server, then, need to attach the parentType as "ComputeResource"
				my $key;        my $value;
				my $parent_is_esx = 0;
				my $parent;
				while ( ($key, $value) = each (%$rp_hash) ) {
					if ($key eq $current_vapp_moref) {
						$parent = $value;
					}
				}

				while ( ($key, $value) = each (%$rp_hash) ) {
					if ($key eq $parent) {
						if ($value eq $current_esx_parent_moref) {
							$parent_is_esx = 1;
							last;
						}
					}
				}

				if ($parent_is_esx) {
					$vapp_xmlelem->appendTextChild('parentType', "ComputeResource");
				} else {
					$vapp_xmlelem->appendTextChild('parentType', $_->get_property('parent')->type);
				}

				$vapps_xmlelem->appendChild($vapp_xmlelem);
			}
		}

	}	# End foreach rp_views

	$server_xmlelem->appendChild($vapps_xmlelem);

	debug_message('end get_va_information');

}   # End 'get_va_information' sub-routine


# Attach only the root resource pool called "Resources"
sub get_only_root_pool_information {

	my $xml_target = shift;
	my $server_xmlelem = shift;
	my $current_esx_parent_moref = shift;
	my $rp_views = shift;

	foreach ( @{$rp_views || []} ) {
		if ($_->get_property('parent')->value eq $current_esx_parent_moref) {
			# This is the root Resource Pool "Resources"
			# Attach it with 2 different XML tags - 'resource_pools' and 'virtual_apps'
			my $root_rp_key = $_->get_property('mo_ref.value');
			my $root_rp_name = $_->get_property('name');
			my $root_rp_parent_key = $_->get_property('parent')->value;

			my $rps_xmlelem = $xml_target->createElement('resource_pools');
			my $rp_xmlelem = $xml_target->createElement('resource_pool');
			$rp_xmlelem->appendTextChild('key', $root_rp_key);
			$rp_xmlelem->appendTextChild('name', $root_rp_name);
			$rp_xmlelem->appendTextChild('parentKey', $root_rp_parent_key);
			$rp_xmlelem->appendTextChild('parentType', 'ComputeResourceDefault');
			$rps_xmlelem->appendChild($rp_xmlelem);
			$server_xmlelem->appendChild($rps_xmlelem);

			# Now attach the "virtual_apps" XML element
			my $vapps_xmlelem = $xml_target->createElement('virtual_apps');
			my $vapp_xmlelem = $xml_target->createElement('virtual_app');
			$vapp_xmlelem->appendTextChild('key', $root_rp_key);
			$vapp_xmlelem->appendTextChild('name', $root_rp_name);
			$vapp_xmlelem->appendTextChild('parentKey', $root_rp_parent_key);
			$vapp_xmlelem->appendTextChild('parentType', 'ComputeResourceDefault');
			$vapps_xmlelem->appendChild($vapp_xmlelem);
			$server_xmlelem->appendChild($vapps_xmlelem);

			last;   # Attached 'Resources'. Good. Exit.
		}
	}

}   # End 'get_only_root_pool_information' sub-routine


# Helper sub-routine to find whether a RP/vApp belongs to the ESX server with MoRef "$current_esx_moref"
sub find_my_parent;
sub find_my_parent {

	my ($rp_hash, $current_rp_moref, $current_esx_moref, $current_esx_parent_moref, $found) = @_;

	my $parent_moref;
	my $key; 	my $value;
	my $value_found = 0;
	my @keys = keys %$rp_hash;

	foreach (@keys) {
		if ($_ eq $current_rp_moref) {
			while (($key, $value) = each(%$rp_hash)) {
				if ($key eq $current_rp_moref) {
					$value_found = 1;
					$parent_moref = $value;
				}
			}
			if ( ($parent_moref eq $current_esx_moref) or ($parent_moref eq $current_esx_parent_moref) ) {
				$$found = 1;
				return;
			}
		}
	}

	if ($value_found) {
		find_my_parent($rp_hash, $parent_moref, $current_esx_moref, $current_esx_parent_moref, $found);
	}
}

# Add datastores related information to the XML
sub attach_datastore_information {

	my $xml_target = shift;
	my $server_xmlelem = shift;
	my $datastores = shift;

	my $datastores_xmlelem = $xml_target->createElement('datastores');

	foreach (@$datastores) {
		my $current_ds_name = $_->get_property('name');
		print STDERR "Looking at datastore - $current_ds_name\n";
		print STDERR "Accessible - " .$_->get_property('summary.accessible'). "\n";
		if ($_->get_property('summary.accessible')) {

			if ( ($current_ds_name =~ '^Unitrends_Instant_Recovery') or
				($current_ds_name =~ '^Unitrends_WIR') ) {
				# Hide VMware and Windows instant recovery datastores
				# since we don't use them for backup/restore.
				print STDERR "Not returning datastore $current_ds_name\n";
				next;
			}

			$current_ds_name = translate_name($current_ds_name, 1);

			my $datastore_xmlelem = $xml_target->createElement('datastore');
			append_to_xml($datastore_xmlelem, 'name', $current_ds_name);
			append_to_xml($datastore_xmlelem, 'capacity', $_->get_property('summary.capacity'));
			append_to_xml($datastore_xmlelem, 'freeSpace', $_->get_property('summary.freeSpace'));
			$datastores_xmlelem->appendChild($datastore_xmlelem);

		} else {
			print STDERR "Datastore $current_ds_name is not accessible. Skipping it.\n";
		}
	}
	
	$server_xmlelem->addChild($datastores_xmlelem);

}   # End 'attach_datastore_information' sub-routine

sub attach_disk_info {

	my $devices = shift;
	my $vm_name = shift;
	my $vm_xmlelem = shift;
	my $xml_target = shift;

	my $disks_xmlelem = $xml_target->createElement('disks');
	foreach(@$devices) {
		if (ref($_) eq 'VirtualDisk') {
			if (not defined ($_->backing)) {
				print STDERR "Skipping disk info for disk " .$_->key.
				" for VM $vm_name\n";
				next;
			}
			my $disk_xmlelem = $xml_target->createElement('disk');

			append_if_def($disk_xmlelem, 'key', $_->{'key'});
			append_if_def($disk_xmlelem, 'capacityInKB', $_->{'capacityInKB'});
			append_if_def($disk_xmlelem, 'fileName', $_->backing->{'fileName'});
			append_if_def($disk_xmlelem, 'uuid', $_->backing->{'uuid'});
			append_if_def($disk_xmlelem, 'diskMode', $_->backing->{'diskMode'});
			append_if_def($disk_xmlelem, 'compatibilityMode', $_->backing->{'compatibilityMode'});
			append_if_def($disk_xmlelem, 'lunUuid', $_->backing->{'lunUuid'});

			$disks_xmlelem->appendChild($disk_xmlelem);
		}
	}
	$vm_xmlelem->appendChild($disks_xmlelem);

}

sub get_disk_information ($$$$) {
	# This sub-routine only retrieves nformation for all the disks for a single VM

	my $vmuuid = shift;
	my $is_template = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;

	debug_message('begin get_disk_information');

	my $vm_moref;
	if ($is_template eq "NO") {
		$vm_moref = VProtect::Backup::get_moref_by_uuid($vmuuid, $xmldoc, $xmlroot);
	} else {
		$vm_moref = VProtect::Backup::get_template_moref_by_uuid($vmuuid, $xmldoc, $xmlroot);
	}
	my $vm = Vim::get_view(mo_ref => $vm_moref);

	my $vm_x = $xmldoc->createElement('virtual_machine');
	my $vm_disk = $xmldoc->createElement('disks');

	if (defined ($vm->config)) {
		my $name = $vm->config->name;
		$name =~ s/%5c/\\/g;
		$name =~ s/%2f/\//g;
		append_if_def($vm_x, 'name', $name);
		append_if_def($vm_x, 'uuid', $vm->config->uuid);

		my $devices = $vm->config->hardware->device;
		foreach(@$devices) {
			if (ref($_) eq ('VirtualDisk')) {
				my $disk_x = $xmldoc->createElement('disk');

				append_if_def($disk_x, 'key', $_->{'key'});
				append_if_def($disk_x, 'capacityInKB', $_->{'capacityInKB'});

				if (not defined($_->{'backing'})) {
					die "Could not identify backing for disk: " . ref $_->{'backing'} . "\n";
				}

				#if (not defined($_->backing->{'uuid'})) {
				#       my $msg = "The virtual disk with filename" . $_->backing->{'filename'} .
				#       " does not have a UUID assigned to it.\n" .
				#       "Please assign a UUID using the command " .
				#       "\"vmkfstools -J setuuid <path_to_vm>/vm_name.vmdk\" on your ESX host.  (See VMware KB 1033370 for more info.)\n" .
				#       "Restart the management agents for this change to take effect. (http://kb.vmware.com/kb/1003490)\n" .
				#       "key " . $_->{'key'} . "\n";
				#       die $msg;
				#}
				#append_or_die($disk_x, 'uuid', $_->backing->{'uuid'});

				append_if_def($disk_x, 'fileName', $_->backing->{'fileName'});
				append_if_def($disk_x, 'uuid', $_->backing->{'uuid'});
				append_if_def($disk_x, 'diskMode', $_->backing->{'diskMode'});
				append_if_def($disk_x, 'compatibilityMode', $_->backing->{'compatibilityMode'});
				append_if_def($disk_x, 'lunUuid', $_->backing->{'lunUuid'});

				$vm_disk->appendChild($disk_x);
			}
		}
	}

	$vm_x->appendChild($vm_disk);
	$xmlroot->appendChild($vm_x);

	debug_message('end get_disk_information');
}

sub get_network_information ($) {

	my $xml_target = shift;

	debug_message('beginning of get_network_information');

	debug_message('vim_get_hosts');
	my $esx_server_views = Vim::find_entity_views(view_type => "HostSystem",
						properties => [ 'summary', 'network' ] );
	debug_message('end vim_get_hosts');

	my $vpinfo_xmlelem = xml_get_or_create($xml_target, 'vprotect_info');

	foreach ( @{$esx_server_views || []} ) {
		my $server_xmlelem = $xml_target->createElement('esx_server');
		print STDERR "Looking at ESX server - " .$_->summary->config->name. "\n";

		$server_xmlelem->appendTextChild('name', $_->summary->config->name);
		$server_xmlelem->appendTextChild('uuid', $_->summary->hardware->uuid);

		my $networks = $_->network;
		my $networks_xmlelem = $xml_target->createElement('networks');

		foreach (@$networks) {
			my $nw_view;
			my $dvs_view;
			my $uplink_found = 0;

			if ($_->type eq 'Network') {
				$nw_view = Vim::get_view( mo_ref => $_,
							properties => [ 'summary.accessible',
									'summary.name' ]);
			} elsif ($_->type eq 'DistributedVirtualPortgroup') {
				$nw_view = Vim::get_view( mo_ref => $_,
							properties => [ 'summary.accessible',
									'summary.name',
									'config.distributedVirtualSwitch',
									'tag' ]);
				$dvs_view = Vim::get_view( mo_ref => $nw_view->{'config.distributedVirtualSwitch'},
							properties => [ 'name' ] );
			} else {
				print STDERR "Unknown network type\n";
			}

			if ($nw_view->{'summary.accessible'} eq 'true') {
				my $nw_xmlelem = $xml_target->createElement('network');
				if( defined($dvs_view) ) {
					my $tag = $nw_view->{'tag'};
					if (defined $tag) {
						foreach(@$tag) {
							if ($_->key eq 'SYSTEM/DVS.UPLINKPG') {
								$uplink_found = 1;
							}
						}
					}
					if ($uplink_found eq 1) {
						# This is an Uplink Port. Don't return it's information
						next;
					}
					my $dv_name = $nw_view->{'summary.name'} . " (" . $dvs_view->{'name'} . ")";
					append_to_xml($nw_xmlelem, 'name', $dv_name);
				} else {
					append_to_xml($nw_xmlelem, 'name', $nw_view->{'summary.name'});
				}
				$networks_xmlelem->appendChild($nw_xmlelem);
			}
		}

		$server_xmlelem->addChild($networks_xmlelem);

		$vpinfo_xmlelem->appendChild($server_xmlelem);
	}

	$xmlroot->appendChild($vpinfo_xmlelem);

	debug_message('end of get_network_information');

}


# Helper sub-routine to convert the command 'get_all_XXXXX' to a
# container type that is accepted by VMware API
sub find_container_type {

	my $container = shift;

	my $container_type;

	if ($container eq "get_all_datacenters") {
		$container_type = "Datacenter";
	} elsif ($container eq "get_all_esx_clusters") {
		$container_type = "ClusterComputeResource";
	} elsif ($container eq "get_all_esx_servers") {
		$container_type = "HostSystem";
	} elsif ($container eq "get_all_datastores") {
		$container_type = "Datastore";
	} elsif ( ($container eq "get_all_resource_pools") or
		($container eq "get_all_vapps") ) {
		$container_type = "ResourcePool";
	} elsif ($container eq "get_all_folders") {
		$container_type = "Folder";
	} else {
		debug_message("Unknown type of container command - $container");
		die "Unknown command to find containers\n";
	}

	return $container_type;

}

# Helper sub-routine to find and return views to a particular type of container
sub get_container_views {

	my $container_type = shift;
	my $container_views;

	# The 'find_entity_views' calls below can be time consuming based on the number
	# of entities of $container_type that are present in the inventory as well as
	# the number of 'properties' that are retrieved
	debug_message("Begin API to find views for $container_type");
	if ($container_type eq "Datacenter") {
		# Datacenters don't have a 'summary' property
		$container_views = Vim::find_entity_views(view_type => $container_type,
							properties => [ 'name' ] );
	} elsif ($container_type eq "ResourcePool") {
		# Need 'parent' property to identify if it is the default 'Resources' or not
		$container_views = Vim::find_entity_views(view_type => $container_type,
							properties => [ 'name',
									'parent',
									'summary' ] );
	} elsif ($container_type eq "Folder") {
		# For 'Folder', browse through the Datacenters first, collect information
		# related to the 'default' folders and then proceed
		$container_views = Vim::find_entity_views(view_type => "Datacenter",
							properties => [ 'name',
									'datastoreFolder',
									'hostFolder',
									'networkFolder',
									'vmFolder' ] );
	} else {
		$container_views = Vim::find_entity_views(view_type => $container_type,
							properties => [ 'name',
									'summary' ] );
	}
	debug_message("End API to find views for $container_type");

	return $container_views;
}

# Helper sub-routine that will populate the default_morefs array
sub get_default_morefs {

	my ($default_morefs, $container_views) = @_;

	# Retrieve the default 'Datacenters' Folder's MoRef that
	# exists at the vCenter level
	debug_message("Getting service content object");
	my $content = vim_get_service_content();
	debug_message("Got service content object");

	my $root_folder_moref = $content->rootFolder->value;
	push(@$default_morefs, $root_folder_moref);
		
	# First, build a list of the 'default' Folders
	foreach ( @{$container_views || []} ) {
		# Collect all default folders for all Datacenters
		my $default_datastore_folder_moref = $_->get_property('datastoreFolder')->value;
		my $default_host_folder_moref = $_->get_property('hostFolder')->value;
		my $default_network_folder_moref = $_->get_property('networkFolder')->value;
		my $default_vm_folder_moref = $_->get_property('vmFolder')->value;

		# Push all these values in the default_morefs array
		push(@$default_morefs, $default_datastore_folder_moref);
		push(@$default_morefs, $default_host_folder_moref);
		push(@$default_morefs, $default_network_folder_moref);
		push(@$default_morefs, $default_vm_folder_moref);
	}

}

# Sub-routine to get all entities of a particular container type
# Eg: Get all the Datacenters belonging to a vCenter
sub get_inventory_info ($$) {

	my $xml_target = shift;
	my $container = shift;

	debug_message("Container type is - $container");

	my $container_type;
	my $container_views;

	my @default_morefs;	# These default objects ALWAYS exist for a vCenter
	my $default_folder = 0;

	my $return_resource_pool = 0;
	my $return_vapp = 0;

	# Decide with what view_type to call VMware API to retrieve appropriate information
	$container_type = find_container_type($container);
	debug_message("Container type is - $container_type");

	if ($container eq "get_all_resource_pools") {
		$return_resource_pool = 1;
	} elsif ($container eq "get_all_vapps") {
		$return_vapp = 1;
	} else {
		debug_message("Not retrieving information for Resource Pools or vApps");
	}

	# Now call the VMware API to retrieve appropriate view for the container
	$container_views = get_container_views($container_type);
	if (not defined $container_views) {
		die "Unable to retrieve information for $container_type.\n";
	}

	if ($container_type eq "Folder") {
		# Collect the MoRefs of the default folders first
		get_default_morefs(\@default_morefs, $container_views);
		debug_message("Default MoRefs are - @default_morefs");

		# Got the required information related to default MoRefs
		# Now, browse through ALL Folders
		debug_message("Begin API to find views for $container_type");
		$container_views = Vim::find_entity_views(view_type => $container_type,
							properties => [ 'name' ] );
		debug_message("Begin API to find views for $container_type");
	}

	my $vpinfo_xmlelem = xml_get_or_create($xml_target, 'vprotect_info');

	if ($container eq "get_all_vapps") {
		$vpinfo_xmlelem->appendTextChild('container_type', 'VirtualApp');
	} else {
		$vpinfo_xmlelem->appendTextChild('container_type', $container_type);
	}

	# Main loop to collect and return the required information
	foreach ( @{$container_views || []} ) {

		if ($container_type eq "HostSystem") {
			my $esx_available = check_esx_availability($_);
			if ($esx_available) {
				debug_message("This ESX is accessible.");
			} else {
				debug_message(" This ESX is currently not accessible. Skipping it.");
				next;
			}
		}

		my $container_name = $_->get_property('name');
		my $container_moref = $_->get_property('mo_ref.value');

		if ( (not defined $container_name) or (not defined $container_moref) ) {
			debug_message("Cannot retrieve name and/or moref for this object. Skipping it.\n");
			next;
		}

		debug_message("Current container is - $container_name | MoRef is - $container_moref");

		if ($container_type eq "ResourcePool") {
			my $parent_type = $_->get_property('parent')->type;
			if ( ($parent_type eq "ComputeResource") or ($parent_type eq "ClusterComputeResource") ) {
				debug_message(" Skipping $container_name, since it is the ESX itself");
				next;
			}

			if ( ((not defined $_->get_property('summary.instanceUuid')) and ($return_vapp)) or
				((defined $_->get_property('summary.instanceUuid')) and ($return_resource_pool)) ) {
				debug_message(" Skipping $container_name, since returning RPs/vApps");
				next;
			}
		}

		if ($container_type eq "Folder") {
			foreach(@default_morefs) {
				if ($_ eq $container_moref) {
					$default_folder = 1;
					last;
				}
			}
			if ($default_folder) {
				$default_folder = 0;
				debug_message(" Skipping folder $container_name, since it is 'default'");
				next;
			}
		}

		if ($container_type eq "Datastore") {
			my $accessible = $_->get_property('summary.accessible');
			if ($accessible == 0) {
				debug_message(" Skipping datastore $container_name, since it is not accessible");
				next;
			}
			if ( ($container_name =~ '^Unitrends_Instant_Recovery') or
				($container_name =~ '^Unitrends_WIR') ) {
				debug_message(" Skipping datastore $container_name, since it was created for WIR/IR");
				next;
			}
		}

		$container_name = translate_name($container_name, 1);

		# This container needs to be attached to the XML
		my $container_xmlelem = $xml_target->createElement('container');
		$container_xmlelem->appendTextChild('name', $container_name);
		$vpinfo_xmlelem->appendChild($container_xmlelem);

	} # End foreach $container_views

	$xmlroot->appendChild($vpinfo_xmlelem);

} # End get_inventory_info sub-routine

######################################################################################################################################################################

# Helper sub-routine to attach a VM's name to the XML
sub attach_vm_info {

	my ($vm_name, $vm_uuid, $xml_target, $vms_xmlelem) = @_;

	# Before attaching a VM's name to the XML, translate the
	# appropriate characters
	$vm_name = translate_name($vm_name, 1);

	my $vm_xmlelem = $xml_target->createElement('virtual_machine');
	$vm_xmlelem->appendTextChild('name', $vm_name);
	$vm_xmlelem->appendTextChild('uuid', $vm_uuid);
	$vms_xmlelem->appendChild($vm_xmlelem);

}

# Helper (recursive) sub-routine to find all the VMs belonging to a Folder
sub find_folder_vms ($$$$);
sub find_folder_vms ($$$$) {

	my $folder_object = shift;
	my $xml_target = shift;
	my $vms_xmlelem = shift;
	my $folder_hash = shift;

	my $folder_name = $folder_object->name;

	my $folder_moref = $folder_object->get_property('mo_ref.value');
	my $accessed = $folder_hash->{$folder_moref};
	if ($accessed) {
		# Information related to this entity has already been attached
		debug_message("   Information related to $folder_name with MoRef $folder_moref has already been returned");
		return;
	} else {
		$folder_hash->{$folder_moref} = 1;	# Mark as accessed/traversed
		if (defined ($folder_object->childEntity)) {
			debug_message("  Returning information for $folder_name with MoRef $folder_moref");
		} else {
			debug_message("   Folder $folder_name is an empty Folder. Returning.");
			next;
		}
	}

	# First attach all the VMs that belong directly to this Folder
	foreach ( @{$folder_object->childEntity} ) {

		if ($_->type eq "VirtualMachine") {
			my $curr_elem_view = vim_get_view($_, "name", "config.instanceUuid");
			if (not defined $curr_elem_view) {
				debug_message("   Unable to get_view information for a VM/Folder. Skipping.");
				next;
			}
		
			# Attach the information for this VM to the XML
			attach_vm_info($curr_elem_view->name, $curr_elem_view->{'config.instanceUuid'}, $xml_target, $vms_xmlelem);
		} elsif ($_->type eq "Folder") {
			# Nested Folder. Traverse it.
			my $curr_folder_view = vim_get_view($_, "name", "childEntity");
			if (not defined $curr_folder_view) {
				debug_message("   Unable to get_view information for a Folder. Skipping.");
				next;
			}

			# Recursive call
			find_folder_vms($curr_folder_view, $xml_target, $vms_xmlelem, $folder_hash);
		} elsif ($_->type eq "Datacenter") {
			# Datacenter belongs to this Folder
			my $datacenter_view = vim_get_view($_, "vmFolder");
			if (not defined $datacenter_view) {
				debug_message("   Unable to get_view information for a Datacenter. Skipping.");
				next;
			}

			my $vm_folder_moref = $datacenter_view->vmFolder;
			# Recursive call
			find_vms_in_folder($vm_folder_moref, $xml_target, $vms_xmlelem);
		} else {
			my $type = $_->type;
			debug_message("   ***Unknown type - $type. Skipping it.***");
		}

	}

}

# Sub-routine to return all VMs belonging to Folder/s named $folder_name
# This will also return all the VMs belonging to a Folder which is nested under $folder_name
sub get_folder_vms ($$$) {

	my $xml_target = shift;
	my $vpinfo_xmlelem = shift;
	my $folder_name = shift;

	my @default_morefs;     # These default objects ALWAYS exist for a vCenter
	my $default_folder = 0;

	my $container_type = "Folder";
	my $temp_folder_views = get_container_views($container_type);
	if (not defined $temp_folder_views) {
                die "Unable to retrieve information for Folder $folder_name.\n";
        }
	get_default_morefs(\@default_morefs, $temp_folder_views);
	#debug_message(" Default MoRefs are - @default_morefs");

	debug_message(" Begin vim get folder views");
	my $folder_views = Vim::find_entity_views( view_type => "Folder",
						properties => [ 'name',
								'childEntity' ] );
	debug_message(" End vim get folder views");

	if (not defined $folder_views) {
		die "Unable to retrieve information for Folder $folder_name.\n";
	}

	my $vms_xmlelem = $xml_target->createElement('virtual_machines');

	my %folder_hash;	# Build a hash of all Folders' MoRefs
	foreach ( @{$folder_views || []} ) {
		my $folder_moref_value = $_->get_property('mo_ref.value');
		$folder_hash{$folder_moref_value} = 0;
	}

	foreach ( @{$folder_views || []} ) {
		my $current_folder_name = $_->name;
		my $current_folder_moref_value = $_->get_property('mo_ref.value');
		foreach(@default_morefs) {
			if ($_ eq $current_folder_moref_value) {
				$default_folder = 1;
				last;
			}
		}
		if ($default_folder) {
			$default_folder = 0;
			$folder_hash{$current_folder_moref_value} = 1;
			debug_message("  Skipping folder $current_folder_name, since it is 'default'");
			next;
		}
		if ($folder_hash{$current_folder_moref_value} eq 1) {
			debug_message("  Folder $current_folder_name with MoRef $current_folder_moref_value has already been accessed. Moving.");
			next;
		}

		debug_message(" Folder name is $current_folder_name");

		if ($current_folder_name eq $folder_name) {
			debug_message("  Folder name matches ! Looking for VMs under this");
			my $current_folder_moref = $_->get_property('mo_ref');
			find_folder_vms($_, $xml_target, $vms_xmlelem, \%folder_hash);
		} else {
			debug_message("  Folder name doesn't match. Skipping.");
		}
	}

	$vpinfo_xmlelem->appendChild($vms_xmlelem);

	$xml_target->appendChild($vms_xmlelem);

} # End get_folder_vms sub-routine

# Sub-routine to return all VMs belonging to Datastore named $datastore_name
sub get_datastore_vms ($$$) {

	my $xml_target = shift;
	my $vpinfo_xmlelem = shift;
	my $datastore_name = shift;

	debug_message(" Begin vim get datastore views");
	my $datastore_views = Vim::find_entity_views( view_type => "Datastore",
						properties => [ 'name',
								'vm',
								'summary' ] );
	debug_message(" End vim get datastore views");

	if (not defined $datastore_views) {
		die "Unable to retrieve information for Datastore $datastore_name.\n";
	}

        my $vms_xmlelem = $xml_target->createElement('virtual_machines');

	foreach ( @{$datastore_views || []} ) {
		my $current_datastore_name = $_->name;

		if ($current_datastore_name eq $datastore_name) {
			my $accessible = $_->get_property('summary.accessible');
			if ($accessible == 0) {
				debug_message(" Skipping datastore $current_datastore_name, since it is not accessible");
				next;
			}
			if ( ($current_datastore_name =~ '^Unitrends_Instant_Recovery') or
				($current_datastore_name =~ '^Unitrends_WIR') ) {
				debug_message(" Skipping datastore $current_datastore_name, since it was created for WIR/IR");
				next;
			}
	
			debug_message(" Returning information for Datastore - $current_datastore_name");

			my $vms = $_->vm;
			foreach( @{$vms} ) {
				my $curr_elem_view = vim_get_view($_, "name", "config.instanceUuid");
				if (not defined $curr_elem_view) {
					debug_message(" Unable to retrieve view information for a VM. Skipping.");
					next;
				}

				# Attach the information for this VM to the XML
				attach_vm_info($curr_elem_view->name, $curr_elem_view->{'config.instanceUuid'}, $xml_target, $vms_xmlelem);
			} # End foreach $vms
		}

	} # End foreach $datastore_views

	$vpinfo_xmlelem->appendChild($vms_xmlelem);

        $xml_target->appendChild($vms_xmlelem);

} # End get_datastore_vms sub-routine

# Helper (recursive) sub-routine to find all the VMs belonging to a ResourcePool/vApp
# and it's child ResourcePools/vApps (if any)
sub find_vms_in_resource_pools ($$$$);
sub find_vms_in_resource_pools ($$$$) {

	my $rp_object = shift;
	my $vms_xmlelem = shift;
	my $xml_target = shift;
	my $rp_hash = shift;

	my $rp_name = $rp_object->name;

	my $rp_moref = $rp_object->get_property('mo_ref.value');
	my $accessed = $rp_hash->{$rp_moref};
	if ($accessed) {
		# Information related to this entity has already been attached
		debug_message("  Information related to $rp_name with MoRef $rp_moref has already been returned");
		return;
	} else {
		debug_message(" Returning information for $rp_name with MoRef $rp_moref");
		$rp_hash->{$rp_moref} = 1;	# Mark it as accessed
	}

	# First attach all the VMs that belong directly to this ResourcePool
	my $vms = $rp_object->vm;
	if (not defined $vms) {
		# No VMs belong directly to this ResourcePool.
		# But, there might be VMs belonging any nested ResourcePools (if any).
		# So, continue... Don't return
		debug_message("  $rp_name doesn't have any VMs that belong directly to it.")
	} else {
		foreach( @{$vms} ) {
			my $curr_elem_view = vim_get_view($_, "name", "config.instanceUuid");
			if (not defined $curr_elem_view) {
				debug_message(" Unable to retrieve view information for a VM. Skipping.");
				next;
			}

			# Attach the information for this VM to the XML
			attach_vm_info($curr_elem_view->name, $curr_elem_view->{'config.instanceUuid'}, $xml_target, $vms_xmlelem);
		}
	}

	# If there are other ResourcePools defined here, find the VMs inside
	# those and attach those VMs as well to the XML
	my $rps = $rp_object->resourcePool;
	if (not defined $rps) {
		# If a ResourcePool doesn't have further nested ResourcePools,
		# it means there are no further VMs down that path. Return.
		debug_message("   $rp_name doesn't have any nested Resource Pools. Not traversing further.");
		return;
	}
	foreach ( @{$rps} ) {
		my $curr_elem_view = vim_get_view($_, "name", "vm", "resourcePool");
		if (not defined $curr_elem_view) {
			debug_message(" Unable to get a view to find information for a RP. Skipping.");
			next;
		}

		# Recursive call
		find_vms_in_resource_pools($curr_elem_view, $vms_xmlelem, $xml_target, $rp_hash);
	}

} # End find_vms_in_resource_pools sub-routine

# Sub-routine to return all VMs belonging to RP/vApp named $container_name
# This will also return all the VMs belonging to a RP/vApp which is nested under $container_name
sub get_rp_vapp_vms ($$$$) {

	my $xml_target = shift;
	my $vpinfo_xmlelem = shift;
	my $container_type = shift;
	my $container_name = shift;

	debug_message(" Begin vim get resource pool views");
	my $rp_views = Vim::find_entity_views( view_type => $container_type,
						properties => [ 'name',
								'vm',
								'resourcePool',
								'summary' ] );
	debug_message(" End vim get resource pool views");

	if (not defined $rp_views) {
		die "Unable to retrieve information for $container_type named - $container_name\n";
	}

	my $vms_xmlelem = $xml_target->createElement('virtual_machines');

	# Build a RP hash indicating whether that ResourcePool has already been
	# accessed/looked at/traversed or not
	# "rp_hash" is of the form/pair -- <RP/vApp MoRef> <===> <0>
	my %rp_hash;
	foreach ( @{$rp_views || []} ) {
		my $rp_moref_value = $_->get_property('mo_ref.value');
		$rp_hash{$rp_moref_value} = 0;	# '0' indicates not traversed/accessed
	}

	foreach ( @{$rp_views || []} ) {
		my $current_rp_name = $_->name;
		debug_message("Name of current RP is $current_rp_name");
		my $current_rp_moref_value = $_->get_property('mo_ref.value');

		if ( ((defined ($_->get_property('summary.instanceUuid'))) and ($container_type eq "ResourcePool")) or
			( (not defined ($_->get_property('summary.instanceUuid'))) and ($container_type eq "VirtualApp")) ) {
			debug_message(" Skipping $current_rp_name since returning $container_type");
			$rp_hash{$current_rp_moref_value} = 1;	# Mark it as accessed
			next;
		}

		if ($current_rp_name eq $container_name) {
			if ($rp_hash{$current_rp_moref_value} eq 0) {
				find_vms_in_resource_pools($_, $vms_xmlelem, $xml_target, \%rp_hash);
			} else {
				debug_message(" Skipping $current_rp_name, since it has already been accessed.");
			}
		}
	}
	
	$vpinfo_xmlelem->appendChild($vms_xmlelem);

        $xml_target->appendChild($vms_xmlelem);

}

# Sub-routine to return all VMs belonging to ESX Server named $esx_server_name
sub get_esx_server_vms ($$$) {

	my $xml_target = shift;
	my $vpinfo_xmlelem = shift;
	my $esx_server_name = shift;

	debug_message(" Begin vim get esx server views");
	my $esx_server_views = Vim::find_entity_views( view_type => "HostSystem",
							properties => [ 'name',
									'vm',
									'summary' ] );
	debug_message(" End vim get esx server views");

	if (not defined $esx_server_views) {
		die "Unable to retrieve information for ESX Server named - $esx_server_name\n";
	}

	my $number_of_esx_servers = scalar(@{$esx_server_views});
        debug_message("Total number of ESX Servers = $number_of_esx_servers");

	my $vms_xmlelem = $xml_target->createElement('virtual_machines');

	foreach ( @{$esx_server_views || []} ) {
		# Traverse through ALL the ESX servers. But, attach information only for the ESX server $esx_cluster_name
		my $current_esx_server_name = $_->name;
		debug_message("Current ESX server name is - $current_esx_server_name");

		my $esx_available = check_esx_availability($_);
		if ($esx_available) {
			debug_message("This ESX is accessible.");
		} else {
			debug_message(" This ESX is currently not accessible. Skipping it.");
			next;
		}

		if ($current_esx_server_name eq $esx_server_name) {
			my $vms = $_->vm;
			debug_message("     Begin for loop for VMs");
			foreach( @{$vms} ) {
				#my $curr_elem_view = vim_get_view($_, "name", "summary.runtime.connectionState", "datastore", "config.ftInfo", "config.extraConfig");
				my $curr_elem_view = vim_get_view($_, "name", "config.instanceUuid");
				if (not defined $curr_elem_view) {
					debug_message(" Unable to retrieve view information for a VM. Skipping.");
					next;
				}

				# Attach the information for this VM to the XML
				attach_vm_info($curr_elem_view->name, $curr_elem_view->{'config.instanceUuid'}, $xml_target, $vms_xmlelem);
			}
			debug_message("     End for loop for VMs");
		}
	}

	$vpinfo_xmlelem->appendChild($vms_xmlelem);

        $xml_target->appendChild($vms_xmlelem);

} # End get_esx_server_vms sub-routine

# Sub-routine to return all VMs belonging to ESX Cluster/s named $esx_cluster_name
sub get_esx_cluster_vms ($$$) {

	my $xml_target = shift;
	my $vpinfo_xmlelem = shift;
	my $esx_cluster_name = shift;

	debug_message(" Begin vim get cluster views");
	my $esx_cluster_views = Vim::find_entity_views( view_type => "ClusterComputeResource",
							properties => [ 'name',
									'resourcePool' ] );
	debug_message(" End vim get cluster views");

	if (not defined $esx_cluster_views) {
		die "Unable to retrieve information for ESX Cluster/s named - $esx_cluster_name.\n";
	}

	my $number_of_esx_clusters = scalar(@{$esx_cluster_views});
	debug_message("Total number of ESX Clusters = $number_of_esx_clusters");

	if ($number_of_esx_clusters eq 0) {
		debug_message("No ESX Clusters are present in the inventory.\n");
		return;
	}

	my $vms_xmlelem = $xml_target->createElement('virtual_machines');

	foreach ( @{$esx_cluster_views || []} ) {
		# Traverse through ALL the ESX clusters. But, attach information only for the cluster $esx_cluster_name
		my $current_cluster_name = $_->name;
		debug_message("Current ESX Cluster name is - $current_cluster_name");

		if ($current_cluster_name eq $esx_cluster_name) {
			# Return VMs for this Cluster
			my $rp_view = vim_get_view($_->resourcePool, "vm", "resourcePool");
			if(not defined $rp_view) {
				debug_message(" Unable to retrieve view information for Cluster $current_cluster_name. Skipping.");
				next;
			}

			# First attach all the VMs that directly belong to this Cluster
			my $vms = $rp_view->vm;
			foreach( @{$vms} ) {
				my $curr_elem_view = vim_get_view($_, "name", "config.instanceUuid");
				if (not defined $curr_elem_view) {
					debug_message(" Unable to retrieve view information for a VM. Skipping.");
					next;
				}

				# Attach the information for this VM to the XML
				attach_vm_info($curr_elem_view->name, $curr_elem_view->{'config.instanceUuid'}, $xml_target, $vms_xmlelem);
			} # End foreach $vms

			# Now, attach all the VMs that belong to the child/nested ResourcePools of this Cluster
			my %rp_hash;	# Dummy hash
			my $rps = $rp_view->resourcePool;
			foreach ( @{$rps} ) {
				my $curr_elem_view = vim_get_view($_, "name", "vm", "resourcePool");
				if (not defined $curr_elem_view) {
					debug_message(" Unable to get a view to find information for a RP. Skipping.");
					next;
				}

				# Recursively find the VMs, RPs and the VMs under those RPs
				find_vms_in_resource_pools($curr_elem_view, $vms_xmlelem, $xml_target, \%rp_hash);
			} # End foreach $rps

		} # End if condition. Proceed to next cluster.
		
	} # End foreach $esx_cluster_views

	$vpinfo_xmlelem->appendChild($vms_xmlelem);

        $xml_target->appendChild($vms_xmlelem);

} # End get_esx_cluster_vms sub-routine

# Helper (recursive) sub-routine to find all the VMs belonging to a single Datacenter
sub find_vms_in_folder ($$$);
sub find_vms_in_folder ($$$) {

	my $folder_moref = shift;
	my $xml_target = shift;
	my $vms_xmlelem = shift;

	my $folder_view = vim_get_view($folder_moref, "name", "childEntity");
	if (not defined $folder_view) {
		debug_message("  Unable to get information for a Folder.");
		die "Unable to retrieve information for a Datacenter.\n";
	}

	my $moref_type = $folder_moref->type;
	debug_message(" MoRef type is $moref_type");

	my $folder_name = $folder_view->name;
	if (not defined $folder_view->childEntity) {
		# Empty Folder
		debug_message("  No VMs belong to Folder $folder_name. Moving to next.");
		return;
	} else {
		debug_message(" Returning VM and Folder information for Folder $folder_name");
	}

	my %rp_hash;	# Dummy hash
	for ( @{$folder_view->childEntity} ) {
		# Traverse the children and find the VMs, Folders and vApps
		my $curr_elem_view;
		my $entity_type = $_->type;

		if ( ($entity_type eq "VirtualMachine") or ($entity_type eq "Folder") ) {

			if ($entity_type eq "VirtualMachine") {
				$curr_elem_view = vim_get_view($_, "name", "config.instanceUuid");
				if (not defined $curr_elem_view) {
					debug_message("  Unable to retrieve view information for a $entity_type. Skipping.");
					next;
				}

				# Attach the information for this VM to the XML
				attach_vm_info($curr_elem_view->name, $curr_elem_view->{'config.instanceUuid'}, $xml_target, $vms_xmlelem);
			} else {
				# Nested Folder. Traverse it (Recursive call).
				find_vms_in_folder($_, $xml_target, $vms_xmlelem);
			}

		} elsif ($entity_type eq "VirtualApp") {
			$curr_elem_view = vim_get_view($_, "name", "vm", "resourcePool");
			if (not defined $curr_elem_view) {
				debug_message("  Unable to retrieve view information for a VirtualApp. Skipping.");
				next;
			}

			# vApp can have nested Resource Pools as well as other vApps. Traverse it.
			find_vms_in_resource_pools($curr_elem_view, $vms_xmlelem, $xml_target, \%rp_hash);

		} else {
			debug_message("   ***$entity_type - is an unknown type. Skipping.***");
		}
	}

} # End find_vms_in_folder sub-routine

# Sub-routine to return all VMs belonging to Datacenter/s named $datacenter_name
sub get_datacenter_vms ($$$) {

	my $xml_target = shift;
	my $vpinfo_xmlelem = shift;
	my $datacenter_name = shift;

	debug_message(" Begin vim get datacenter views");
	my $datacenter_views = Vim::find_entity_views(view_type => "Datacenter",
						properties => [ 'name',
								'vmFolder' ] );
	debug_message(" End vim get datacenter views");

	if (not defined $datacenter_views) {
		die "Unable to retrieve information for Datacenter/s named - $datacenter_name\n";
	}

	my $number_of_datacenters = scalar(@{$datacenter_views});
        debug_message("Total number of Datacenters = $number_of_datacenters");

	my $vms_xmlelem = $xml_target->createElement('virtual_machines');
	foreach ( @{$datacenter_views || []} ) {
		# Traverse through ALL the Datacenters since there can be Datacenters with the same name
		my $current_dc_name = $_->name;
		debug_message("Current Datacenter name is $current_dc_name");

		if ($current_dc_name eq $datacenter_name) {
			my $vm_folder_moref = $_->vmFolder;
			# Recursively find the VMs existing in all Folders and sub-Folders
			find_vms_in_folder($vm_folder_moref, $xml_target, $vms_xmlelem);
		}
	}

	$vpinfo_xmlelem->appendChild($vms_xmlelem);

	$xml_target->appendChild($vms_xmlelem);

}

# Sub-routine to get all the VMs belonging to $container_name of type $container_type
# There can be more than one container of type $container_type which is/are named $container_name
sub get_container_vms ($$$) {

	my $xml_target = shift;
	my $container_type = shift;
	my $container_name = shift;

	debug_message("Begin get_container_vms");
	debug_message("Returning VMs belonging to - $container_type - named - $container_name");

	$container_name = translate_name($container_name, 0);
	debug_message("Container name after decoding is --- $container_name");

	my $vpinfo_xmlelem = xml_get_or_create($xml_target, 'vprotect_info');

	if ($container_type eq "Datacenter") {
		# Get VMs belonging to Datacenter/s named $container_name
		get_datacenter_vms($xml_target, $vpinfo_xmlelem, $container_name);
	} elsif ($container_type eq "ClusterComputeResource") {
		# Get VMs belonging to ESX Cluster/s named $container_name
		get_esx_cluster_vms($xml_target, $vpinfo_xmlelem, $container_name);
	} elsif ($container_type eq "HostSystem") {
		# Get VMs belonging to ESX Server/s named $container_name
		get_esx_server_vms($xml_target, $vpinfo_xmlelem, $container_name);
	} elsif ( ($container_type eq "ResourcePool") or
			($container_type eq "VirtualApp") ) {
		# Get VMs belonging to Resource Pool/s / vApp/s named $container_name
		get_rp_vapp_vms($xml_target, $vpinfo_xmlelem, $container_type, $container_name);
	} elsif ($container_type eq "Datastore") {
		# Get VMs belonging to Datastore/s named $container_name
		get_datastore_vms($xml_target, $vpinfo_xmlelem, $container_name);
	} elsif ($container_type eq "Folder") {
		# Get VMs belonging to Folder/s named $container_name
		get_folder_vms($xml_target, $vpinfo_xmlelem, $container_name);
	} else {
		die "Container Type $container_type is not supported.\n";
	}
	
	$xmlroot->appendChild($vpinfo_xmlelem);
	
	debug_message("End get_container_vms");

}


sub output_xml_error ($$) {
	my $xmltarget = shift;
	my $errstring = shift;

	chomp $errstring;

	my $xmlelem = $xmltarget->createElement('vprotect_error');
	$xmlelem->appendTextChild('error_message', $errstring);
	$xmlroot->appendChild($xmlelem);

	return 0;
}

sub output_xml_warning ($$) {
	my $xmltarget = shift;
	my $warnstring = shift;

	chomp $warnstring;

	my $xmlelem = $xmltarget->createElement('vprotect_warning');
	$xmlelem->appendTextChild('warning_message', $warnstring);
	$xmlroot->appendChild($xmlelem);
}

sub promote_next_reason ($$) {
	my $xmltarget = shift;
	my $reason = shift;

	my $xmlelem = $xmltarget->createElement('promote_next_backup');
	$xmlelem->appendText('1');
	$xmlroot->appendChild($xmlelem);

	my $xmlelem_reason = $xmltarget->createElement('promote_next_backup_reason');
	$xmlelem_reason->appendText($reason);
	$xmlroot->appendChild($xmlelem_reason);
}

sub promote_next ($) {
	my $xmltarget = shift;

	my $xmlelem = $xmltarget->createElement('promote_next_backup');
	$xmlelem->appendText('1');
	$xmlroot->appendChild($xmlelem);

	my $xmlelem_reason = $xmltarget->createElement('promote_next_backup_reason');
	$xmlelem_reason->appendText('A VM configuration change has been detected.');
	$xmlroot->appendChild($xmlelem_reason);
}

sub translate_name ($$) {

	my $orig_name = shift;
	my $decode = shift;

	my $translated_name = $orig_name;

	if ($decode) {
		# Encode:
		#	%5c to \
		#	%2f to /
		#	%25 to %
		$translated_name =~ s/%5c/\\/g;
		$translated_name =~ s/%2f/\//g;
		$translated_name =~ s/%25/\%/g;
	} else {
		# Decode:
		#	% to %25
		#	/ to %2f
		#	\ to %5c
		$translated_name =~ s/\%/%25/g;
		$translated_name =~ s/\//%2f/g;
		$translated_name =~ s/\\/%5c/g;
	}

	return $translated_name;

}

##### end functions


##### utility functions 

sub validate ($) {
	my $xmldoc = shift;
	my $valid = 1;

	my $mode_opts = 0;

	if (Opts::option_is_set('backup')) {
		$mode_opts += 1;
	}
	if (Opts::option_is_set('bpcmd')) {
		$mode_opts += 1;
	}
	if (Opts::option_is_set('ensure_ds_empty')) {
		$mode_opts += 1;
	}
	if (Opts::option_is_set('restore')) {
		$mode_opts += 1;
	}
	if (Opts::option_is_set('instant_recovery')) {
		$mode_opts += 1;
	}
	if (Opts::option_is_set('removesnap')) {
		$mode_opts += 1;
	}
	if (Opts::option_is_set('remove_ir_datastore')) {
		$mode_opts += 1;
	}
	if (Opts::option_is_set('disklist')) {
		$mode_opts += 1;
	}
	if (Opts::option_is_set('wir')) {
		$mode_opts += 1;
	}
	if (Opts::option_is_set('ensure_vm_removed')) {
		$mode_opts += 1;
	}
	if (Opts::option_is_set('delete_vm')) {
		$mode_opts += 1;
	}
	if (Opts::option_is_set('containertype')) {
		$mode_opts += 1;
	}

	if ($mode_opts == 0) {
		output_xml_error($xmldoc, 'No mode specified: choose from backup, restore, instant_recovery, bpcmd, removesnap, remove_ir_datastore, ensure_ds_empty.');
		print $xmldoc->toString(1);
		exit(2);
	}

	if ($mode_opts > 1) {
		output_xml_error($xmldoc, 'An invalid combination of arguments was provided.  (Multiple mode options.)');
		print $xmldoc->toString(1);
		exit(2);
	}

	if (defined(Opts::get_option('restore'))) {
		if (!Opts::option_is_set('hostname') ||
			!Opts::option_is_set('newvmname') ||
			!Opts::option_is_set('datastore_name')) {
				output_xml_error($xmldoc, 'Arguments "datastore_name", "hostname", and "newvmname" are required for restore.');
				print $xmldoc->toString(1);
				exit(2);
			}
	}

	if (defined(Opts::get_option('instant_recovery'))) {
		if (!Opts::option_is_set('server') ||
			!Opts::option_is_set('hostname') ||
			!Opts::option_is_set('newvmname') ||
			!Opts::option_is_set('datastore_name') ||
			!Opts::option_is_set('ir_datastore_host') ||
			!Opts::option_is_set('ir_datastore_path') ||
			!Opts::option_is_set('ir_disk_map_path') ||
			!Opts::option_is_set('xml_filename')) {
			output_xml_error($xmldoc, 'Arguments "server", "hostname", "newvmname", "ir_datastore_host", "ir_datastore_path", "ir_disk_map_path", and "xml_filename" are required for instant recovery.');
			print $xmldoc->toString(1);
			exit(2);
		}
	}

	if (!Opts::option_is_set('username')) {
		output_xml_error($xmldoc, "Username not supplied.");
		print $xmldoc->toString(1);
		exit(2);
	}

	if (!Opts::option_is_set('password')) {
		output_xml_error($xmldoc, "Password not supplied.");
		print $xmldoc->toString(1);
		exit(2);
	}

	return $valid;
}

##### end utility functions


exit main();
